from typing import Any, Iterator, List, Optional, Union, cast

from tomlkit import TOMLDocument, items
from tomlkit.container import OutOfOrderTableProxy

from tomlkit_extras._exceptions import InvalidArrayItemError
from tomlkit_extras._typing import (
    AnnotatedContainer,
    BodyContainerItem,
    ContainerComment,
    TOMLHierarchy,
)
from tomlkit_extras._utils import decompose_body_item, get_container_body
from tomlkit_extras.descriptor._descriptor import TOMLDocumentDescriptor
from tomlkit_extras.toml._out_of_order import fix_out_of_order_table
from tomlkit_extras.toml._retrieval import get_attribute_from_toml_source


def _container_has_comments(attribute: Union[TOMLDocument, items.Item]) -> bool:
    """
    Private function to determine if a specific tomlkit type can contain
    annotations.
    """
    return isinstance(attribute, (TOMLDocument, items.Table, items.AoT, items.Array))


def get_comments(
    toml_source: AnnotatedContainer, hierarchy: Optional[TOMLHierarchy] = None
) -> Optional[List[ContainerComment]]:
    """
    Retrieves and returns all comments appearing in the top-level space of a given
    tomklit type. A TOML source of type `AnnotatedContainer` must be passed in.

    If no heirarchy is specified then the search will occur in the TOML source
    passed. Otherwise if a hierarchy is included, then it must be relative to
    the source. The item located at the hierarchy will be retrieved and the
    search will occur within that item.

    Returns a tuple where the first item is the line number where the comment
    is located and the second item is the comment itself.

    Will return None if no comments were found.

    Args:
        toml_source (`AnnotatedContainer`): An `AnnotatedContainer` instance.
        hierarchy (`TOMLHierarchy` | None): None or a `TOMLHierarchy` instance.

    Returns:
        List[`ContainerComment`] | None: None if no comments were found, or a list of
            `ContainerComment` instances.
    """
    if isinstance(toml_source, items.Array) or hierarchy is None:
        if isinstance(toml_source, OutOfOrderTableProxy):
            toml_source = fix_out_of_order_table(table=toml_source)

        attributes = [toml_source]
    else:
        attribute = get_attribute_from_toml_source(
            hierarchy=hierarchy, toml_source=toml_source, fix_order=True
        )
        if not isinstance(attribute, list) or isinstance(attribute, items.AoT):
            attribute = [attribute]

        if not all(_container_has_comments(attribute=attr) for attr in attribute):
            raise ValueError("Attribute is not a structure that can contain comments")

        attributes = cast(
            List[Union[TOMLDocument, items.Table, items.Array]], attribute
        )

    comments: List[ContainerComment] = []
    for attr in attributes:
        document_descriptor = TOMLDocumentDescriptor(
            toml_source=attr, top_level_only=True
        )

        for comment_descriptor in document_descriptor.get_top_level_stylings(
            styling="comment"
        ):
            comments.append((comment_descriptor.line_no, comment_descriptor.style))

    return comments if comments else None


def get_array_field_comment(array: items.Array, array_item: Any) -> Optional[str]:
    """
    Will return the comment associated with an array item appearing within a
    `tomlkit.items.Array` instance. Association in this case means the first comment
    appearing after the item in question, but before any whitespace (a new line).

    Will return None if no comment was found.

    Args:
        array (`tomlkit.items.Array`): A `tomlkit.items.Array` instance.
        array_item (Any): Any type corresponding to an array item.

    Returns:
        str | None: None if no comment was found, or a string comment if found.
    """
    array_items = get_container_body(toml_source=array)
    array_items_iter: Iterator[BodyContainerItem] = iter(array_items)

    seen_first_ws_after_comment: bool = False
    seen_array_item: bool = False
    array_item_comment: Optional[str] = None

    try:
        while (
            not (seen_array_item and seen_first_ws_after_comment)
            and array_item_comment is None
        ):
            _, array_body_item = decompose_body_item(body_item=next(array_items_iter))

            if not seen_array_item:
                seen_array_item = array_body_item == array_item
            elif isinstance(array_body_item, items.Whitespace):
                seen_first_ws_after_comment = "\n" in array_body_item.value
            elif isinstance(array_body_item, items.Comment):
                array_item_comment = array_body_item.trivia.comment
    except StopIteration:
        pass

    if not seen_array_item:
        raise InvalidArrayItemError(
            "Data item does not exist in specified array", array_items
        )

    return array_item_comment



import os
from pathlib import Path
from typing import Optional, Union

import charset_normalizer
import tomlkit
from charset_normalizer import CharsetMatch
from pathvalidate import ValidationError, validate_filepath
from tomlkit import TOMLDocument
from tomlkit.exceptions import ParseError

from tomlkit_extras._exceptions import TOMLConversionError, TOMLDecodingError
from tomlkit_extras._typing import TOMLSourceFile
from tomlkit_extras._utils import from_dict_to_toml_document


def _read_toml(toml_content: str) -> TOMLDocument:
    """
    A private function which converts an instance of a string, that being
    a string representation of a TOML file, into a `tomlkit.TOMLDocument`
    instance.
    """
    try:
        toml_content_parsed: TOMLDocument = tomlkit.parse(toml_content)
        return toml_content_parsed
    except ParseError as e:
        raise TOMLDecodingError("Issue occured when decoding the TOML source content")
    except Exception:
        raise TOMLConversionError(
            "Unexpected issue occured when loading the source from TOML"
        )


def _load_toml(toml_content: Union[str, bytes]) -> TOMLDocument:
    """
    A private function which accepts either a string or bytes instance,
    being a string or bytes representation of a TOML file respectively, into
    a `tomlkit.TOMLDocument` instance.
    """
    if isinstance(toml_content, bytes):
        detected_encoding: Optional[CharsetMatch] = charset_normalizer.from_bytes(
            toml_content
        ).best()

        # Default to utf-8 encoding if encoding was not detected
        toml_encoding: str = "utf-8"

        if detected_encoding is not None:
            toml_encoding = detected_encoding.encoding

        # Decode content and parse into dictionary
        toml_content_decoded: str = toml_content.decode(toml_encoding)
        return _read_toml(toml_content=toml_content_decoded)
    else:
        return _read_toml(toml_content=toml_content)


def load_toml_file(toml_source: TOMLSourceFile) -> TOMLDocument:
    """
    Accepts a string, bytes, bytearray, `Path`, `tomlkit.TOMLDocument`, or
    Dict[str, Any] instance and converts it into a `tomlkit.TOMLDocument`
    instance.

    Args:
        toml_source (`TOMLSourceFile`): A string, bytes, bytearray, Path,
            `tomlkit.TOMLDocument`, or Dict[str, Any] instance.

    Returns:
        `tomlkit.TOMLDocument`: A `tomlkit.TOMLDocument` instance.
    """
    if isinstance(toml_source, (str, Path)):
        if os.path.isfile(toml_source):
            with open(toml_source, mode="rb") as file:
                toml_content = file.read()

            return _load_toml(toml_content=toml_content)

        try:
            toml_source_as_path = Path(toml_source)
            validate_filepath(file_path=toml_source_as_path)
        except ValidationError:
            pass
        else:
            raise FileNotFoundError(
                "If path is passed in as the source, it must link to an existing file"
            )

        if isinstance(toml_source, str):
            return _load_toml(toml_content=toml_source)
        else:
            raise TOMLConversionError(
                "Unexpected issue occured when loading the source from TOML"
            )
    elif isinstance(toml_source, TOMLDocument):
        return toml_source
    elif isinstance(toml_source, dict):
        return from_dict_to_toml_document(dictionary=toml_source)

    # If the source is passed as a bytes object
    elif isinstance(toml_source, bytes):
        return _load_toml(toml_content=toml_source)

    # In the case where the source is passed as a bytearray object
    elif isinstance(toml_source, bytearray):
        toml_source_to_bytes = bytes(toml_source)

        return _load_toml(toml_content=toml_source_to_bytes)
    else:
        raise TypeError(
            "Expected an instance of TOMLSourceFile, but got "
            f"{type(toml_source).__name__}"
        )



from typing import Any, Dict, List, Optional, Tuple, Type, Union, cast

import tomlkit
from tomlkit import TOMLDocument, container, items
from tomlkit.container import OutOfOrderTableProxy

from tomlkit_extras._hierarchy import Hierarchy, standardize_hierarchy
from tomlkit_extras._typing import (
    BodyContainer,
    BodyContainerItem,
    BodyContainerItemDecomposed,
    BodyContainerItems,
    TOMLDictLike,
    TOMLHierarchy,
    TOMLSource,
    TOMLValidReturn,
)

_VALID_TYPES: Tuple[Type[TOMLSource], ...] = (
    TOMLDocument,
    items.Table,
    items.AoT,
    OutOfOrderTableProxy,
)


def safe_unwrap(structure: Union[TOMLDocument, TOMLValidReturn]) -> Any:
    """
    Safely unwraps a `tomlkit` object, which is the action of returning
    the underlying value that the `tomlkit` type is wrapping.

    This safe logic is needed because even though most types contain an
    `unwrap` method, there are a few in which it is not implemented.

    Args:
        structure (`tomlkit.TOMLDocument` | `TOMLValidReturn`): A
            `tomlkit.TOMLDocument` or `TOMLValidReturn` instance.

    Returns:
        Any: Any instance, including pritimitive types and others.
    """
    if isinstance(structure, items.Whitespace):
        return structure.value
    elif isinstance(structure, items.Comment):
        return structure.as_string()
    elif isinstance(structure, bool):
        return structure
    else:
        return structure.unwrap()


def contains_out_of_order_tables(toml_source: TOMLSource) -> bool:
    """
    Given a `TOMLSource` instance, will traverse through the structure
    and determine if there are any out-of-order tables represented by
    instances of `tomlkit.container.OutOfOrderTableProxy`.

    Returns a boolean that is True if there are out-of-order tables and
    False if not.

    Args:
        toml_source (`TOMLSource`): A `TOMLSource` instance.

    Returns:
        bool: A boolean indicating whether there are any out-of-order tables
            located in the TOML structure.
    """

    def out_of_order_detect(source: TOMLSource) -> bool:
        # Recursively iterate through dictionary-like tomlkit structures
        if isinstance(source, (TOMLDocument, items.Table)):
            return any(
                out_of_order_detect(source=document_value)
                for _, document_value in source.items()
                # Only recursively traverse through sub-structures if they
                # are tomlkit structures that can contain out-of-order tables
                if isinstance(document_value, _VALID_TYPES)
            )
        # Recursively iterate through list-like tomlkit structures
        # Because items in an array of tables can only be tables, there is
        # no need for a valid type check
        elif isinstance(source, items.AoT):
            return any(out_of_order_detect(source=aot_table) for aot_table in source)
        # Otherwise check if structure is an out-of-order table
        return isinstance(source, OutOfOrderTableProxy)

    if not isinstance(toml_source, _VALID_TYPES):
        raise TypeError(
            f"Expected an instance of TOMLSource, but got {type(toml_source).__name__}"
        )

    return out_of_order_detect(source=toml_source)


def find_comment_line_no(line_no: int, item: items.Item) -> Optional[int]:
    """
    Given a line number and an `tomlkit.items.Item` instance, will calculate
    the line number in which the comment associated with the item lies on.

    If there is no comment found then will return None.

    Args:
        line_no (int): The line number in which the item is located.
        item (`tomlkit.items.Item`): A `tomlkit.items.Item` instance.

    Returns:
        int | None: An integer line number where the comment is located, if
            there is no comment then returns None.
    """
    comment_position: Optional[int]

    if not item.trivia.comment:
        comment_position = None
    else:
        ws_before_comment: str = item.trivia.indent + item.trivia.comment_ws
        num_newlines = ws_before_comment.count("\n")
        comment_position = line_no + num_newlines

    return comment_position


def from_dict_to_toml_document(dictionary: Dict[str, Any]) -> TOMLDocument:
    """
    Converts a dictionary into a `tomlkit.TOMLDocument` instance.

    This function takes a dictionary with string keys and values of any type
    and converts it into a `tomlkit.TOMLDocument`, which is a structured
    representation of TOML data.

    Args:
        dictionary (Dict[str, Any]): A dictionary with keys as strings and values
            being of any type.

    Returns:
        `tomlkit.TOMLDocument`: A `tomlkit.TOMLDocument` instance.
    """
    toml_document: TOMLDocument = tomlkit.document()
    for document_key, document_value in dictionary.items():
        toml_document.add(key=document_key, item=document_value)

    return toml_document


def convert_to_tomlkit_item(value: Any) -> items.Item:
    """
    Converts an instance of any type into an `tomlkit.items.Item` instance.

    If the argument is already of type `tomlkit.items.Item`, then the conversion
    is skipped and the input is automatically returned.

    Args:
        value (Any): An instance of any type.

    Returns:
        `tomlkit.items.Item`: A `tomlkit.items.Item` instance.
    """
    if not isinstance(value, items.Item):
        value_as_toml_item = tomlkit.item(value=value)
    else:
        value_as_toml_item = value
    return value_as_toml_item


def create_array_of_tables(
    tables: List[Union[items.Table, Dict[str, Any]]]
) -> items.AoT:
    """
    Converts a list of `tomlkit.items.Table` instances or list of dictionaries,
    each with keys as strings and values being of any type, into a
    `tomlkit.items.AoT` instance.

    Args:
        tables (List[`tomlkit.items.Table` | Dict[str, Any]): A list of
            `tomlkit.items.Table` instances or list of dictionaries, each with
            keys as strings and values being of any type

    Returns:
        `tomlkit.items.AoT`: A `tomlkit.items.AoT` instance.
    """
    array_of_tables: items.AoT = tomlkit.aot()
    for table in tables:
        array_of_tables.append(table)
    return array_of_tables


def create_table(fields: Dict[str, Any]) -> items.Table:
    """
    Given a dictionary, will return an `items.Table` instance where the
    fields are are the key-value pairs.

    Args:
        fields (Dict[str, Any]): A general dictionary where the keys are
            strings and values can be of any type.

    Returns:
        `items.Table`: Returns an `items.Table` instance.
    """
    table: items.Table = tomlkit.table()
    table.update(fields)
    return table


def create_inline_table(fields: Dict[str, Any]) -> items.InlineTable:
    """
    Given a dictionary, will return an `items.InlineTable` instance where the
    fields are are the key-value pairs.

    Args:
        fields (Dict[str, Any]): A general dictionary where the keys are
            strings and values can be of any type.

    Returns:
        `items.InlineTable`: Returns an `items.InlineTable` instance.
    """
    inline_table: items.InlineTable = tomlkit.inline_table()
    inline_table.update(fields)
    return inline_table


def create_array(entries: List[Any]) -> items.Array:
    """
    Given a list, will return an `items.Array` instance where the items are
    the entries from the input list.

    Args:
        entries (List[Any]): A general list where the items can be of any type.

    Returns:
        `items.Array`: Returns an `items.Array` instance.
    """
    array: items.Array = tomlkit.array()
    array.extend(entries)
    return array


def create_toml_document(hierarchy: TOMLHierarchy, value: Any) -> TOMLDocument:
    """
    Given a hierarchy of string or `Hierarchy` type, and a value being an
    instance of any type, will create a `tomlkit.TOMLDocument` instance inserting
    the value at the hierarchy, specified. Thus, creating a `tomlkit.TOMLDocument`
    instance around the value.

    If the value inserted is not already and instance of `tomlkit.items.Item`,
    will automatically convert into a `tomlkit.items.Item` instance.

    Args:
        hierarchy (`TOMLHierarchy`): A `TOMLHierarchy` instance.
        value (Any): An instance of any type.

    Returns:
        `tomlkit.TOMLDocument`: A `tomlkit.TOMLDocument` instance.
    """
    hierarchy_obj: Hierarchy = standardize_hierarchy(hierarchy=hierarchy)
    source: TOMLDocument = tomlkit.document()

    current_source: Union[items.Table, TOMLDocument] = source
    for table in hierarchy_obj.hierarchy:
        current_source[table] = tomlkit.table()
        current_source = cast(items.Table, current_source[table])

    current_source[hierarchy_obj.attribute] = convert_to_tomlkit_item(value=value)

    return source


def _partial_clear_dict_like_toml_item(toml_source: TOMLDictLike) -> None:
    """
    A private function that deletes all key-value pairs appearing in
    a `TOMLDictLike` instance.
    """
    keys = list(toml_source.keys())
    for key in keys:
        dict.__delitem__(toml_source, key)


def complete_clear_toml_document(toml_document: TOMLDocument) -> None:
    """
    Completely resets a `tomlkit.TOMLDocument` instance, including
    deleting all key-value pairs and all private attributes storing
    data.

    Args:
        toml_document (`tomlkit.TOMLDocument`): A `tomlkit.TOMLDocument` instance.
    """
    _partial_clear_dict_like_toml_item(toml_source=toml_document)

    # Reset private attributes that store elements within document
    toml_document._map = {}
    toml_document._body = []
    toml_document._parsed = False
    toml_document._table_keys = []


def complete_clear_out_of_order_table(table: OutOfOrderTableProxy) -> None:
    """
    Completely resets a `tomlkit.container.OutOfOrderTableProxy` instance,
    including deleting all key-value pairs and all private attributes storing
    data.

    Args:
        table (`tomlkit.container.OutOfOrderTableProxy`):
            A `tomlkit.container.OutOfOrderTableProxy` instance.
    """
    _partial_clear_dict_like_toml_item(toml_source=table)

    # Reset private attributes that store elements within table
    table._container = container.Container()
    table._internal_container = container.Container(True)
    table._tables = []
    table._tables_map = {}


def complete_clear_tables(table: Union[items.Table, items.InlineTable]) -> None:
    """
    Completely resets a `tomlkit.items.Table` or `tomlkit.items.InlineTable`
    instance, including deleting all key-value pairs and all private attributes
    storing data.

    Args:
        table (`tomlkit.items.Table` | `tomlkit.items.InlineTable`): A `tomlkit.items.Table`
            or `tomlkit.items.InlineTable` instance.
    """
    _partial_clear_dict_like_toml_item(toml_source=table)

    # Reset private attributes that store elements within table
    table._value = container.Container()


def complete_clear_array(array: items.Array) -> None:
    """
    Completely resets a `tomlkit.items.Array` instance.

    Args:
        array (`tomlkit.items.Array`): A `tomlkit.items.Array` instance.
    """
    array.clear()


def _reorganize_array(array: items.Array) -> BodyContainerItems:
    """
    A private function which reorganizes a `tomlkit.items.Array` instance and
    returns a `BodyContainerItems` type. This function effectively applies a
    standardization on the body of an array.
    """
    array_body_items: BodyContainerItems = []

    for array_item_group in array._value:
        for array_item in array_item_group:
            array_body_items.append((None, array_item))

    return array_body_items


def get_container_body(toml_source: BodyContainer) -> BodyContainerItems:
    """
    Retrieves the core elements, making up the body of a `BodyContainer` type,
    and returns a `BodyContainerItems` type.

    Args:
        toml_source (`BodyContainer`): A `BodyContainer` instance.

    Returns:
        `BodyContainerItems`: A `BodyContainerItems` instance.
    """
    if isinstance(toml_source, (items.Table, items.InlineTable)):
        table_body_items = toml_source.value.body
    elif isinstance(toml_source, items.Array):
        table_body_items = _reorganize_array(array=toml_source)
    elif isinstance(toml_source, OutOfOrderTableProxy):
        table_body_items = toml_source._container.body
    elif isinstance(toml_source, TOMLDocument):
        table_body_items = toml_source.body
    else:
        raise ValueError("Type is not a valid container-like structure")
    return table_body_items


def decompose_body_item(body_item: BodyContainerItem) -> BodyContainerItemDecomposed:
    """
    Decomposes an item, from the body of a `BodyContainer` type, being
    of type `BodyContainerItem`, and returns a `BodyContainerItemDecomposed` type.

    Args:
        body_item (`BodyContainerItem`): A `BodyContainerItem` instance.

    Returns:
        `BodyContainerItemDecomposed`: A `BodyContainerItemDecomposed` instance.
    """
    raw_key = body_item[0]
    item_key: Optional[str] = (
        raw_key.as_string().strip() if raw_key is not None else None
    )
    toml_item: items.Item = body_item[1]
    return item_key, toml_item



import copy
from typing import Any, Iterator, List, Optional, cast

from tomlkit import TOMLDocument, items
from tomlkit.container import OutOfOrderTableProxy

from tomlkit_extras._typing import TOMLSource


def _fix_of_out_of_order_table_chain(
    current_table: items.Table, update_key: str, update_table: items.Table
) -> None:
    """
    A private function which will iterate through a chain of tables and fixes any
    out-of-order tables that are encountered. Within `tomlkit`, these out-of-order
    tables are represented by `tomlkit.container.OutOfOrderTableProxy` types.

    This manipulations are done in-place.
    """
    if update_key not in current_table:
        current_table[update_key] = update_table
    else:
        update_from_table: items.Table = _find_child_table(
            table_value=current_table[update_key]
        )

        if update_from_table.is_super_table():
            new_current_table = update_table
            new_update_table = update_from_table
            current_table[update_key] = update_table
        else:
            new_current_table = update_from_table
            new_update_table = update_table

        for table_key, table_value in new_update_table.items():
            child_table = _find_child_table(table_value=table_value)

            _fix_of_out_of_order_table_chain(
                current_table=new_current_table,
                update_key=table_key,
                update_table=child_table,
            )


def _find_child_table(table_value: Any) -> items.Table:
    """
    A private function which runs the main function to fix any out-of-order-tables
    that are encountered. Otherwise it returns the argument value.
    """
    if isinstance(table_value, OutOfOrderTableProxy):
        child_table = fix_out_of_order_table(table=table_value)
    else:
        child_table = table_value
    return child_table


def fix_out_of_order_table(table: OutOfOrderTableProxy) -> items.Table:
    """
    Given an out-of-order table, represented by a `tomlkit.container.OutOfOrderTableProxy`
    instance, will fix the order of the table and return a `tomlkit.items.Table`
    instance.

    Out-of-order tables occur when a descendant of a hierarchy appears above
    that hierarchy within a TOML file. In these cases, instead of returning a
    `tomlkit.items.Table` instance as would normally be done, a
    `tomlkit.container.OutOfOrderTableProxy` is generated. While the ordering in
    the TOML file is maintained, it poses some issues. Since it is not a sub-class
    of `tomlkit.items.Item`, it therefore does not contain a comment or other info.

    This simple function was built to parse and fix any of those issues by
    re-ordering the table. It navigates through the entire structure, so if there
    are any nested out-of-order tables, these will be fixed as well.

    Args:
        table (`tomlkit.container.OutOfOrderTableProxy`): A
            `tomlkit.container.OutOfOrderTableProxy` instance.

    Returns:
        `tomlkit.items.Table`: A `tomlkit.items.Table` instance.
    """
    component_tables = cast(List[items.Table], copy.deepcopy(table._tables))

    table_w_shortest_name: Optional[items.Table] = None
    parent_table: Optional[items.Table] = None

    component_tables_iter: Iterator[items.Table] = iter(component_tables)

    try:
        while parent_table is None:
            co_table: items.Table = next(component_tables_iter)

            if not co_table.is_super_table():
                parent_table = co_table
            elif co_table.name is not None:
                if table_w_shortest_name is None:
                    table_w_shortest_name = co_table
                elif table_w_shortest_name.name is not None and len(
                    co_table.name
                ) < len(table_w_shortest_name.name):
                    table_w_shortest_name = co_table
    except StopIteration:
        parent_table = table_w_shortest_name

    parent_table = cast(items.Table, parent_table)
    component_tables.remove(parent_table)

    for component_table in component_tables:
        current_table = parent_table

        for table_key, table_value in component_table.items():
            child_table: items.Table = _find_child_table(table_value=table_value)

            _fix_of_out_of_order_table_chain(
                current_table=current_table,
                update_key=table_key,
                update_table=child_table,
            )

    return parent_table


def fix_out_of_order_tables(toml_source: TOMLSource) -> None:
    """
    Fixes all out-of-order tables, represented by a `tomlkit.container.OutOfOrderTableProxy`
    instances, appearing in a `TOMLSource` instance. The re-ordering an manipluations are
    done in-place and no type is returned.

    Args:
        toml_source (`TOMLSource`): A `TOMLSource` instance.
    """
    if isinstance(toml_source, (items.Table, TOMLDocument)):
        for table_key, table_value in toml_source.items():
            if isinstance(table_value, OutOfOrderTableProxy):
                toml_source[table_key] = fix_out_of_order_table(table=table_value)
            elif isinstance(table_value, items.Table):
                fix_out_of_order_tables(toml_source=table_value)
    elif isinstance(toml_source, items.AoT):
        for aot_table in toml_source:
            fix_out_of_order_tables(toml_source=aot_table)
    else:
        raise TypeError(
            f"Expected an instance of TOMLSource, but got {type(toml_source).__name__}"
        )



from abc import ABC, abstractmethod
from typing import Any, List, Optional, Set, Union

from tomlkit_extras._hierarchy import Hierarchy
from tomlkit_extras._typing import BodyContainerItems, Retrieval, TOMLFieldSource
from tomlkit_extras._utils import decompose_body_item, safe_unwrap

# ==============================================================================
# General Base Error Class for all Errors
# ==============================================================================


class BaseTOMLError(Exception):
    """
    Base error class for all errors thrown in `tomlkit_extras`.

    Attributes:
        message (str): The string error message.
    """

    def __init__(self, message: str) -> None:
        self.message = message

    def __str__(self) -> str:
        return repr(self)

    def __repr__(self) -> str:
        return self.message


# ==============================================================================
# TOML Reading Errors
# ==============================================================================


class TOMLReadError(BaseTOMLError):
    """
    Base error class for those related to reading in TOML files.

    Inherits attributes from `BaseTOMLError`:
    - `message`
    """

    pass


class TOMLDecodingError(TOMLReadError):
    """
    Decoding error occurring when loading a TOML configuration file.

    Inherits attributes from `BaseTOMLError`:
    - `message`
    """

    pass


class TOMLConversionError(TOMLReadError):
    """
    General TOML conversion error when reading in TOML files.

    Inherits attributes from `BaseTOMLError`:
    - `message`
    """

    pass


# ==============================================================================
# Invalid Hierarchy Errors
# ==============================================================================


class InvalidTOMLStructureError(ABC, BaseTOMLError):
    """
    Abstract error class for those related to querying a hierarchy from a
    tomlkit structure.

    Inherits attributes from `BaseTOMLError`:
    - `message`

    Attributes:
        hierarchy (`Hierarchy`): A representation of the hierarchy passed
            in as an argument which does not exist.
    """

    def __init__(self, message: str, hierarchy: Hierarchy) -> None:
        super().__init__(message=message)
        self._hierarchy_obj = hierarchy

    @property
    def hierarchy(self) -> str:
        """Returns a string representation of the hierarchy that was invalid."""
        return str(self._hierarchy_obj)

    @property
    @abstractmethod
    def closest_hierarchy(self) -> Optional[str]:
        """Returns the longest ancestor hierarchy that exists in the TOML file."""
        pass


class InvalidHierarchyError(InvalidTOMLStructureError):
    """
    Error occurring when a hierarchy does not exist in set of expected
    hierarchies.

    Inherits attributes from `InvalidTOMLStructureError`:
    - `message`
    - `hierarchy`

    Attributes:
        hierarchies (Set[str]): All hierarchies appearing in the space of
            the TOML file that is being queried. For example, if retrieving
            from a field in the top-level document space, then the hierarchies
            would represent all existing fields in that space.
        closest_hierarchy (str | None): The longest ancestor hierarchy of
            the invalid hierarchy that exists in the TOML file
    """

    def __init__(
        self, message: str, hierarchy: Hierarchy, hierarchies: Set[str]
    ) -> None:
        super().__init__(message=message, hierarchy=hierarchy)
        self.hierarchies = hierarchies

    @property
    def closest_hierarchy(self) -> Optional[str]:
        """Returns the longest ancestor hierarchy that exists in the TOML file."""
        return self._hierarchy_obj.longest_ancestor_hierarchy(
            hierarchies=self.hierarchies
        )


class InvalidFieldError(InvalidTOMLStructureError):
    """
    Error occurring when a field does not exist in set of expected fields.

    Inherits attributes from `InvalidTOMLStructureError`:
    - `message`
    - `hierarchy`

    Attributes:
        field (str): The string field name, corresponding to the last level
            of the hierarchy.
        existing_fields (Set[str]): All fields that exist at the same level
            where the field was expected to be located.
        closest_hierarchy (str | None): The longest ancestor hierarchy of
            the invalid hierarchy that exists in the TOML file
    """

    def __init__(self, message: str, hierarchy: Hierarchy, fields: Set[str]) -> None:
        super().__init__(message=message, hierarchy=hierarchy)
        self.existing_fields = fields
        self.field: str = self._hierarchy_obj.attribute

    @property
    def closest_hierarchy(self) -> str:
        """Returns the longest ancestor hierarchy that exists in the TOML file."""
        return self._hierarchy_obj.base_hierarchy_str


class InvalidTableError(InvalidTOMLStructureError):
    """
    Error occurring when a table does not exist in a set of expected tables.
    This is thrown only when attempting to retrieve a table that exists
    within an array-of-tables.

    Inherits attributes from `InvalidTOMLStructureError`:
    - `message`
    - `hierarchy`

    Attributes:
        table (str): The string table name, corresponding to the last level
            of the hierarchy.
        existing_tables (Set[str]): All hierarchies that correspond to tables
            existing within arrays-of-tables.
        closest_hierarchy (str | None): The longest ancestor hierarchy of
            the invalid hierarchy that exists in the TOML file
    """

    def __init__(self, message: str, hierarchy: Hierarchy, tables: Set[str]) -> None:
        super().__init__(message=message, hierarchy=hierarchy)
        self.existing_tables = tables
        self.table: str = self._hierarchy_obj.attribute

    @property
    def closest_hierarchy(self) -> Optional[str]:
        """Returns the longest ancestor hierarchy that exists in the TOML file."""
        return self._hierarchy_obj.longest_ancestor_hierarchy(
            hierarchies=self.existing_tables
        )


class InvalidArrayOfTablesError(InvalidTOMLStructureError):
    """
    Error occurring when referencing an invalid array of tables.

    Inherits attributes from `InvalidTOMLStructureError`:
    - `message`
    - `hierarchy`

    Attributes:
        arrays (Set[str]): All hierarchies corresponding to existing
            arrays-of-tables in the TOML file.
        closest_hierarchy (str | None): The longest ancestor hierarchy of
            the invalid hierarchy that exists in the TOML file
    """

    def __init__(self, message: str, hierarchy: Hierarchy, arrays: Set[str]) -> None:
        super().__init__(message=message, hierarchy=hierarchy)
        self.arrays = arrays

    @property
    def closest_hierarchy(self) -> Optional[str]:
        """Returns the longest ancestor hierarchy that exists in the TOML file."""
        return self._hierarchy_obj.longest_ancestor_hierarchy(hierarchies=self.arrays)


# ==============================================================================
# Invalid Hierarchy when Modifying Errors (when using TOML utility functions)
# ==============================================================================


class HierarchyModificationError(BaseTOMLError):
    """
    Base error class for those related to modification of a hierarchy in a
    tomlkit structure. Error raised when using the TOML utility functions.

    Inherits attributes from `BaseTOMLError`:
    - `message`
    """

    pass


class InvalidHierarchyDeletionError(HierarchyModificationError):
    """
    Error ocurring when a TOML hierarchy that is to be deleted does not exist.

    Inherits attributes from `HierarchyModificationError`:
    - `message`
    """

    pass


class InvalidHierarchyUpdateError(HierarchyModificationError):
    """
    Error ocurring when a TOML hierarchy that is to be updated does not exist.

    Inherits attributes from `HierarchyModificationError`:
    - `message`
    """

    pass


class InvalidHierarchyRetrievalError(HierarchyModificationError):
    """
    Error ocurring when an object from a TOML hierarchy that is to be retrieved
    does not exist.

    Inherits attributes from `HierarchyModificationError`:
    - `message`
    """

    pass


# ==============================================================================
# Other TOML-related errors
# ==============================================================================


class KeyNotProvidedError(BaseTOMLError):
    """
    Error occuring when inserting into a `tomlkit` object and no string key has
    been provided.

    Inherits attributes from `BaseTOMLError`:
    - `message`
    """

    pass


class NotContainerLikeError(BaseTOMLError):
    """
    Error occuring when an object that is expected to be a container-like
    structure, which can "contain" nested objects, is not.

    Inherits attributes from `BaseTOMLError`:
    - `message`
    """

    pass


class TOMLInsertionError(BaseTOMLError):
    """
    Error occuring when attempting to insert into an object that does not
    support insertion.

    Inherits attributes from `BaseTOMLError`:
    - `message`

    Attributes:
        struct_type (Type[`Retrieval` | `TOMLFieldSource`]): The type of
            the object when the insertion attempt was made.
    """

    def __init__(
        self, message: str, structure: Union[Retrieval, TOMLFieldSource]
    ) -> None:
        super().__init__(message=message)
        self.struct_type = type(structure)


class InvalidStylingError(BaseTOMLError):
    """
    Error occurring when a styling does not exist in set of expected stylings.

    Inherits attributes from `BaseTOMLError`:
    - `message`

    Attributes:
        stylings (Set[str]): All the unique stylings that appear within the
            TOML structure. This is a set that is pre-filtered based on whether
            the query was a comment or whitespace-like string.
    """

    def __init__(self, message: str, stylings: Set[str]) -> None:
        super().__init__(message=message)
        self.stylings = stylings


class InvalidArrayItemError(BaseTOMLError):
    """
    Error occurring when an item expected to exist in an array, does not.

    Inherits attributes from `BaseTOMLError`:
    - `message`

    Attributes:
        array_items (List[Any]): A list of all values appearing in the
            TOML array object.
    """

    def __init__(self, message: str, body: BodyContainerItems) -> None:
        super().__init__(message=message)

        self.array_items: List[Any] = [
            safe_unwrap(structure=decompose_body_item(body_item=item)[1])
            for item in body
        ]



import sys
from pathlib import Path
from typing import Any, Dict, List, Literal, Optional, Tuple, Union

if sys.version_info >= (3, 10):
    from typing import TypeAlias
else:
    from typing_extensions import TypeAlias

from tomlkit import TOMLDocument, items
from tomlkit.container import OutOfOrderTableProxy

from tomlkit_extras._hierarchy import Hierarchy

TOMLSourceFile: TypeAlias = Union[
    str, bytes, bytearray, Path, TOMLDocument, Dict[str, Any]
]

# The return type of get_comments function, returning a tuple where the first item
# is the line number where the comment is located and the second is the comment
ContainerComment: TypeAlias = Tuple[int, str]

# Tomlkit types that are subclasses of dictionaries
TOMLDictLike: TypeAlias = Union[
    TOMLDocument, items.Table, items.InlineTable, OutOfOrderTableProxy
]

# Tomlkit types that are tables
Table: TypeAlias = Union[items.Table, items.InlineTable]

# Tomlkit types that are classified as styles/stylings
Stylings: TypeAlias = Union[items.Whitespace, items.Comment]

# Valid tomlkit return types when traversing a TOMLDocument
TOMLValidReturn: TypeAlias = Union[OutOfOrderTableProxy, items.Item]

# Valid tomlkit return types from the get_attribute_from_toml_source function
Retrieval: TypeAlias = Union[TOMLValidReturn, List[items.Item]]

# Valid hierarchy types in most functions in package
TOMLHierarchy: TypeAlias = Union[str, Hierarchy]

# Various types that have to do with tomlkit types that contain a body of fields,
# tables, and stylings.
BodyContainerItem: TypeAlias = Tuple[Optional[items.Key], items.Item]
BodyContainerItems: TypeAlias = List[BodyContainerItem]
BodyContainerItemDecomposed: TypeAlias = Tuple[Optional[str], items.Item]
BodyContainer: TypeAlias = Union[
    TOMLDocument, items.Table, items.InlineTable, items.Array, OutOfOrderTableProxy
]
BodyContainerInOrder: TypeAlias = Union[
    TOMLDocument, items.Table, items.InlineTable, items.Array
]

# Tomlkit types that can contain key-value pairs
TOMLFieldSource: TypeAlias = Union[
    TOMLDocument, items.Table, items.InlineTable, items.AoT, OutOfOrderTableProxy
]

# Valid input tomlkit types for most TOML related functions in package
TOMLSource: TypeAlias = Union[
    TOMLDocument, items.Table, items.AoT, OutOfOrderTableProxy
]

# Valid input tomlkit types for the TOMLDocumentDescriptor class
DescriptorInput: TypeAlias = Union[TOMLDocument, items.Table, items.AoT, items.Array]

# Tomlkit types that can have comments in the top-level space
AnnotatedContainer: TypeAlias = Union[
    TOMLDocument, items.Table, items.Array, OutOfOrderTableProxy
]

# Any tomlkit type that can contain other objects nested within
ContainerLike: TypeAlias = Union[
    TOMLDocument,
    items.Table,
    items.Array,
    OutOfOrderTableProxy,
    items.AoT,
    items.InlineTable,
]

# Literals identifying the TOML item type for a given descriptor
StyleItem: TypeAlias = Literal["whitespace", "comment"]
TableItem: TypeAlias = Literal["table", "inline-table"]
FieldItem: TypeAlias = Literal["field", "array"]
AoTItem: TypeAlias = Literal["array-of-tables"]
TopLevelItem: TypeAlias = Literal["document", "table", "array-of-tables"]
Item: TypeAlias = Literal[
    "document",
    "field",
    "table",
    "inline-table",
    "super-table",
    "array",
    "array-of-tables",
    "whitespace",
    "comment",
]
ParentItem: TypeAlias = Literal[
    "document", "table", "inline-table", "super-table", "array", "array-of-tables"
]



from __future__ import annotations

from typing import List, Optional, cast

from tomlkit import TOMLDocument, items
from tomlkit.container import OutOfOrderTableProxy

from tomlkit_extras._hierarchy import Hierarchy
from tomlkit_extras._typing import (
    BodyContainerInOrder,
    BodyContainerItems,
    DescriptorInput,
    StyleItem,
    Stylings,
    Table,
    TOMLHierarchy,
    TopLevelItem,
)
from tomlkit_extras._utils import decompose_body_item, get_container_body
from tomlkit_extras.descriptor._descriptors import (
    AoTDescriptor,
    FieldDescriptor,
    StyleDescriptor,
    TableDescriptor,
)
from tomlkit_extras.descriptor._helpers import LineCounter, get_item_type
from tomlkit_extras.descriptor._retriever import DescriptorRetriever
from tomlkit_extras.descriptor._store import DescriptorStore
from tomlkit_extras.descriptor._types import ItemInfo, ItemPosition, TOMLStatistics
from tomlkit_extras.toml._out_of_order import fix_out_of_order_table


class _TOMLParser:
    """
    A private parser class that houses all logic to accurately recursively parse
    through various `tomlkit` types. These types include `tomlkit.items.Array`,
    `tomlkit.items.InlineTable`, `tomlkit.items.Comment`, `tomlkit.items.Whitespace`,
    `tomlkit.items.AoT`, and `tomlkit.items.Table` among others.
    """

    def __init__(
        self,
        line_counter: LineCounter,
        store: DescriptorStore,
        toml_statistics: TOMLStatistics,
        top_level_only: bool,
    ) -> None:
        self._line_counter = line_counter
        self._store = store
        self._toml_statistics = toml_statistics

        self.top_level_only = top_level_only

    def _generate_descriptor_from_aot(self, array: items.AoT, info: ItemInfo) -> None:
        """
        Private method that parses all objects within a `tomlkit.items.AoT`
        instance.

        Initial pre-processing of the array-of-tables occurs, and then the main
        recursive method `_generate_descriptor` is called to continue parsing any
        nested structures.
        """
        array_name = cast(str, array.name)
        hierarchy = Hierarchy.create_hierarchy(
            hierarchy=info.hierarchy, attribute=array_name
        )

        # Generate an AoTDescriptor object to initialize mini-store for nested
        # structures
        array_of_tables = AoTDescriptor(line_no=self._line_counter.line_no, info=info)

        # Append the newly-created AoTDescriptor object to array-of-tables store
        self._store.array_of_tables.append(
            hierarchy=hierarchy, array_of_tables=array_of_tables
        )

        # Iterate through each table in the body of the array and run the
        # main recursive parsing method on the table
        for index, table in enumerate(array.body):
            self._toml_statistics.add_table(table=table)
            table_item_info = ItemInfo.from_parent_type(
                key=array_name,
                hierarchy=info.hierarchy,
                toml_item=table,
                parent_type="array-of-tables",
                from_aot=True,
            )

            # Run the main recursive parsing method on the table
            table_item_info.position = ItemPosition(index + 1, index + 1)
            self._generate_descriptor(container=table, info=table_item_info)

    def _parse_array(self, toml_item: items.Array, info: ItemInfo) -> None:
        """Private method to parse through `items.Array` instances."""
        self._store.update_field_descriptor(item=toml_item, info=info)
        self._generate_descriptor(container=toml_item, info=info)
        self._store.update_array_comment(array=toml_item, info=info)

        # Add array to TOML summary statistics
        self._toml_statistics.add_array(item=toml_item)

        # Add a single line to line counter and update both attribute
        # and container positions
        self._line_counter.add_line()
        info.position.update_positions()

    def _parse_inline_table(self, toml_item: items.InlineTable, info: ItemInfo) -> None:
        """Private method to parse through `items.InlineTable` instances."""
        self._generate_descriptor(container=toml_item, info=info)

        # Add inline-table to TOML summary statistics
        self._toml_statistics.add_inline_table(table=toml_item)

        # Add a single line to line counter and update both attribute
        # and container positions
        self._line_counter.add_line()
        info.position.update_positions()

    def _parse_stylings(self, toml_item: Stylings, info: ItemInfo) -> None:
        """Private method to parse through `items.Whitespace`/`items.Comment` instances."""
        number_of_newlines = toml_item.as_string().count("\n")
        self._store.update_styling(style=toml_item, style_info=info)

        # Add comment to TOML summary statistics
        self._toml_statistics.add_comment(item=toml_item)

        # Add the number of new lines appearing in the styling to
        # the line counter and update only the container position
        self._line_counter.add_lines(lines=number_of_newlines)
        info.position.update_body_position()

    def _parse_array_of_tables(self, toml_item: items.AoT, info: ItemInfo) -> None:
        """Private method to parse through `items.AoT` instances."""
        self._generate_descriptor_from_aot(array=toml_item, info=info)

        # Add array to TOML summary statistics
        self._toml_statistics.add_aot()

        # Update both attribute and container positions
        info.position.update_positions()

    def _parse_table(self, toml_item: items.Table, info: ItemInfo) -> None:
        """Private method to parse through `items.Table` instances."""
        self._generate_descriptor(container=toml_item, info=info)

        # Add table to TOML summary statistics
        self._toml_statistics.add_table(table=toml_item)

        # Update boh attribute and container positions
        info.position.update_positions()

    def _parse_others(
        self, toml_item: items.Item, info: ItemInfo, container: BodyContainerInOrder
    ) -> None:
        """Private method to parse through other `items.Item` instances."""
        if not isinstance(container, items.Array):
            self._store.update_field_descriptor(item=toml_item, info=info)
            if not isinstance(container, items.InlineTable):
                self._line_counter.add_line()

            self._toml_statistics.add_field(item=toml_item)
        info.position.update_positions()

    def _generate_descriptor(
        self, container: BodyContainerInOrder, info: ItemInfo
    ) -> None:
        """
        Private recursive method that traverses an entire `BodyContainerInOrder`
        instance, being a `tomlkit` type `tomlkit.TOMLDocument`, `tomlkit.items.Table`,
        `tomlkit.items.InlineTable`, or `tomlkit.items.Array`.

        During traversal, each field, table, array, and styling (comment and
        whitespace) is parsed and a custom/curated set of data points are collected
        for each item parsed.
        """
        position = ItemPosition.default_position()
        new_hierarchy = Hierarchy.create_hierarchy(
            hierarchy=info.hierarchy, attribute=info.key
        )
        is_non_super_table = (
            isinstance(container, items.Table) and not container.is_super_table()
        )

        # If an tomlkit.items.InlineTable or tomlkit.items.Table, then add
        # a new table to the table store
        if isinstance(container, items.InlineTable) or is_non_super_table:
            self._store.update_table_descriptor(
                hierarchy=new_hierarchy,
                table=cast(Table, container),
                table_info=info,
            )

        # Since an inline table is contained only on a single line, and thus
        # on the same line as the table header, only update the line counter
        # if parsing a tomlkit.TOMLDocument or tomlkit.items.Table instance
        table_body_items: BodyContainerItems = get_container_body(toml_source=container)
        if isinstance(container, TOMLDocument) or is_non_super_table:
            self._line_counter.add_line()

        # Iterate through each item appearing in the body of the tomlkit object
        for toml_body_item in table_body_items:
            item_key, toml_item = decompose_body_item(body_item=toml_body_item)
            toml_item_info = ItemInfo.from_body_item(
                hierarchy=new_hierarchy,
                container_info=info,
                body_item=(item_key, toml_item),
            )

            # Set the position property for the active item
            toml_item_info.position = position

            # If the item is an out-of-order table, then fix and update the
            # item type attribute
            if isinstance(toml_item, OutOfOrderTableProxy):
                toml_item = fix_out_of_order_table(table=toml_item)
                toml_item_info.item_type = "table"

            # If an array is encountered, the function is run recursively since
            # an array can contain stylings and nested tomlkit.items.Item objects
            if isinstance(toml_item, items.Array):
                self._parse_array(toml_item=toml_item, info=toml_item_info)

            # If an inline table is parsed, the function is run recursively since
            # an inline table can contain nested tomlkit.items.Item objects
            elif isinstance(toml_item, items.InlineTable):
                self._parse_inline_table(toml_item=toml_item, info=toml_item_info)

            # If one of the two styling objects are encountered, then the
            # styling is added to the store. No recursive call as stylings
            # cannot contain nested tomlkit objects
            elif isinstance(toml_item, (items.Comment, items.Whitespace)):
                self._parse_stylings(toml_item=toml_item, info=toml_item_info)

            # For an array-of-tables, recursive call is made as arrays can
            # contain any tomlkit object nested within except a tomlkit.TOMLDocument
            elif isinstance(toml_item, items.AoT) and not self.top_level_only:
                self._parse_array_of_tables(toml_item=toml_item, info=toml_item_info)

            # For a non-inline table, make a recursive call
            elif isinstance(toml_item, items.Table) and not self.top_level_only:
                self._parse_table(toml_item=toml_item, info=toml_item_info)

            # Otherwise the object is a generic tomlkit.items.Item
            elif not isinstance(toml_item, (items.Table, items.AoT)):
                self._parse_others(
                    toml_item=toml_item, info=toml_item_info, container=container
                )


class TOMLDocumentDescriptor:
    """
    A class that iterates through, maps out, and collects all relevant
    information for all fields, tables and stylings appearing in a
    `DescriptorInput` instance. A `DescriptorInput` instance is a `tomlkit`
    type of either `tomlkit.TOMLDocument`, `tomlkit.items.Table`,
    `tomlkit.items.AoT`, or `tomlkit.items.Array`.

    Parsing occurs within the constructor. Methods are provided to retrieve
    basic summary statistics about the TOML file, and to extract granular
    information about fields, tables, or stylings appearing at a specific
    hierarchy.

    All out-of-order tables appearing in the TOML file will automatically be
    fixed when parsing. Thus, the line numbers may be innacurate for these
    TOML files.

    Args:
        toml_source (`DescriptorInput`): A `tomlkit` type of either
            `tomlkit.TOMLDocument`, `tomlkit.items.Table`, `tomlkit.items.AoT`,
            or `tomlkit.items.Array`
        top_level_only (bool): A boolean value that indicates whether only the
            top-level space of the `DescriptorInput` structure should be parsed.
            Defaults to False.
    """

    def __init__(
        self, toml_source: DescriptorInput, top_level_only: bool = False
    ) -> None:
        if not isinstance(
            toml_source, (TOMLDocument, items.Table, items.AoT, items.Array)
        ):
            raise TypeError(
                "Expected an instance of DescriptorInput, but got "
                f"{type(toml_source).__name__}"
            )

        self.top_level_only = top_level_only
        self.top_level_type: TopLevelItem = cast(
            TopLevelItem, get_item_type(toml_item=toml_source)
        )
        self.top_level_hierarchy: Optional[str] = (
            toml_source.name
            if isinstance(toml_source, (items.AoT, items.Table))
            else None
        )

        # Tracker for number of lines in TOML
        self._line_counter = LineCounter()

        # Statistics on number of types within TOML source
        self._toml_statistics = TOMLStatistics()

        # Descriptor store
        self._store = DescriptorStore(line_counter=self._line_counter)

        # TOML parser
        self._toml_parser = _TOMLParser(
            line_counter=self._line_counter,
            store=self._store,
            toml_statistics=self._toml_statistics,
            top_level_only=self.top_level_only,
        )

        # Descriptor retriever
        self._retriever = DescriptorRetriever(
            store=self._store,
            top_level_type=self.top_level_type,
            top_level_hierarchy=self.top_level_hierarchy,
        )

        if isinstance(toml_source, (items.Table, items.AoT)):
            update_key = toml_source.name
            assert (
                update_key is not None
            ), "table or array-of-tables must have a string name"
        else:
            update_key = str()

        container_info = ItemInfo.from_parent_type(
            key=update_key, hierarchy=str(), toml_item=toml_source
        )

        # Initialize the main functionality depending on whether the source
        # is an array-of-tables or not
        if isinstance(toml_source, items.AoT):
            self._toml_parser._generate_descriptor_from_aot(
                array=toml_source, info=container_info
            )
        else:
            self._toml_parser._generate_descriptor(
                container=toml_source, info=container_info
            )

        self._line_counter.reset_line_no()

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}"
            f"(type={self.top_level_type!r}, hierarchy={self.top_level_hierarchy!r})"
        )

    @property
    def number_of_tables(self) -> int:
        """
        Returns an integer representing the number of non-inline/non-super tables
        appearing in the TOML file.
        """
        return self._toml_statistics.number_of_tables

    @property
    def number_of_inline_tables(self) -> int:
        """
        Returns an integer representing the number of inline tables appearing in
        the TOML file.
        """
        return self._toml_statistics.number_of_inline_tables

    @property
    def number_of_aots(self) -> int:
        """
        Returns an integer representing the number of array-of-tables appearing
        in the TOML file.
        """
        return self._toml_statistics.number_of_aots

    @property
    def number_of_arrays(self) -> int:
        """
        Returns an integer representing the number of arrays appearing in the
        TOML file.
        """
        return self._toml_statistics.number_of_arrays

    @property
    def number_of_comments(self) -> int:
        """
        Returns an integer representing the number of comments appearing in the
        TOML file.
        """
        return self._toml_statistics.number_of_comments

    @property
    def number_of_fields(self) -> int:
        """
        Returns an integer representing the number of non-array fields appearing
        in the TOML file.
        """
        return self._toml_statistics.number_of_fields

    def get_field_from_aot(self, hierarchy: TOMLHierarchy) -> List[FieldDescriptor]:
        """
        Retrieves all fields from an array-of-tables, where each field is represented
        by a `FieldDescriptor` object, that correspond to a specific hierarchy.

        Args:
            hierarchy (`TOMLHierarchy`) A `TOMLHierarchy` instance.

        Returns:
            List[`FieldDescriptor`]: A list of `FieldDescriptor` instances.
        """
        return self._retriever.get_field_from_aot(hierarchy=hierarchy)

    def get_table_from_aot(self, hierarchy: TOMLHierarchy) -> List[TableDescriptor]:
        """
        Retrieves all tables from an array-of-tables, where each table is represented
        by a `TableDescriptor` object, that correspond to a specific hierarchy.

        Args:
            hierarchy (`TOMLHierarchy`) A `TOMLHierarchy` instance.

        Returns:
            List[`TableDescriptor`]: A list of `TableDescriptor` instances.
        """
        return self._retriever.get_table_from_aot(hierarchy=hierarchy)

    def get_aot(self, hierarchy: TOMLHierarchy) -> List[AoTDescriptor]:
        """
        Retrieves all array-of-tables, where each array is represented
        by a `AoTDescriptor` object, that correspond to a specific hierarchy.

        Args:
            hierarchy (`TOMLHierarchy`) A `TOMLHierarchy` instance.

        Returns:
            List[`AoTDescriptor`]: A list of `AoTDescriptor` instances.
        """
        return self._retriever.get_aot(hierarchy=hierarchy)

    def get_field(self, hierarchy: TOMLHierarchy) -> FieldDescriptor:
        """
        Retrieves a field represented by a `FieldDescriptor` object which
        corresponds to a specific hierarchy.

        Args:
            hierarchy (`TOMLHierarchy`) A `TOMLHierarchy` instance.

        Returns:
            `FieldDescriptor`: A `FieldDescriptor` instance.
        """
        return self._retriever.get_field(hierarchy=hierarchy)

    def get_table(self, hierarchy: TOMLHierarchy) -> TableDescriptor:
        """
        Retrieves a table represented by a `TableDescriptor` object which
        corresponds to a specific hierarchy.

        Args:
            hierarchy (`TOMLHierarchy`) A `TOMLHierarchy` instance.

        Returns:
            `TableDescriptor`: A `TableDescriptor` instance.
        """
        return self._retriever.get_table(hierarchy=hierarchy)

    def get_top_level_stylings(
        self, styling: Optional[StyleItem] = None
    ) -> List[StyleDescriptor]:
        """
        Retrieves all stylings (comments or whitespace) that occur at the
        top-level space of the TOML source.

        If "whitespace" is passed all whitespace stylings will be returned. If
        "comment" is passed all comment stylings will be returned. If it is None,
        then all stylings will be returned.

        Args:
            styling (`StyleItem` | None): A literal that identifies the type of
                styling to retrieve. Can be either "whitespace" or "comment". Is
                optional and defaults to None.

        Returns:
            List[`StyleDescriptor`]: A list of `StyleDescriptor` instances.
        """
        return self._retriever.get_top_level_stylings(styling=styling)

    def get_stylings(
        self, styling: str, hierarchy: Optional[TOMLHierarchy] = None
    ) -> List[StyleDescriptor]:
        """
        Retrieves all stylings corresponding to a specific string representation,
        where each styling is represented by a `StyleDescriptor` object. In
        addition, if the search should be narrowed, a `TOMLHierarchy` object
        can be passed.

        A styling can either be whitespace or comment.

        Args:
            styling (str): A string representation of a comment or whitespace.
            hierarchy (`TOMLHierarchy` | None) A `TOMLHierarchy` instance. Is
                optional and defaults to None.

        Returns:
            List[`StyleDescriptor`]: A list of `StyleDescriptor` instances.
        """
        return self._retriever.get_stylings(styling=styling, hierarchy=hierarchy)



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional, cast

from tomlkit import items

from tomlkit_extras._typing import (
    BodyContainerItemDecomposed,
    DescriptorInput,
    Item,
    ParentItem,
)
from tomlkit_extras.descriptor._helpers import get_item_type


@dataclass
class ItemPosition:
    """
    A class that stores positional information for a specific `tomlkit` object
    while recursively traversing a TOML structure in the `_generate_descriptor`
    method of `_TOMLParser`.

    The attribute position refers to the position of an item amongst all
    other key value pairs (fields, tables) within the containing object. The
    container position is the position of the item amongst all other types,
    including stylings (whitespace, comments), within the containing object.
    """

    attribute: int
    container: int

    @classmethod
    def default_position(cls) -> ItemPosition:
        """
        Class method that generates a default position where the attribute and
        container positions are both one.
        """
        return ItemPosition(attribute=1, container=1)

    def update_position(self) -> None:
        """Add one to the `attribute` positional property."""
        self.attribute += 1

    def update_body_position(self) -> None:
        """Add one to the `container` positional property."""
        self.container += 1

    def update_positions(self) -> None:
        """Add one to the `attribute` and `container` positional properties."""
        self.update_position()
        self.update_body_position()


class ItemInfo:
    """
    A class that stores general information for a specific `tomlkit` object
    while recursively traversing a TOML structure in the `_generate_descriptor`
    method of `_TOMLParser`.
    """

    def __init__(
        self,
        item_type: Item,
        parent_type: Optional[ParentItem],
        key: str,
        hierarchy: str,
        from_aot: bool,
    ) -> None:
        self.item_type = item_type
        self.parent_type = parent_type
        self.key = key
        self.hierarchy = hierarchy
        self.from_aot = from_aot

        self._position: ItemPosition

    @property
    def position(self) -> ItemPosition:
        """Returns the `ItemPosition` object associated with item."""
        return self._position

    @position.setter
    def position(self, position: ItemPosition) -> None:
        """Sets the `ItemPosition` object associated with item."""
        self._position = position

    @classmethod
    def from_parent_type(
        cls,
        key: str,
        hierarchy: str,
        toml_item: DescriptorInput,
        parent_type: Optional[ParentItem] = None,
        from_aot: bool = False,
    ) -> ItemInfo:
        """
        A class method that creates an `ItemInfo` instance from parent
        information.
        """
        item_type = get_item_type(toml_item=toml_item)
        return cls(
            item_type=item_type,
            parent_type=parent_type,
            key=key,
            hierarchy=hierarchy,
            from_aot=from_aot,
        )

    @classmethod
    def from_body_item(
        cls,
        hierarchy: str,
        container_info: ItemInfo,
        body_item: BodyContainerItemDecomposed,
    ) -> ItemInfo:
        """
        A class method that creates an `ItemInfo` instance from the body
        information for an item.
        """
        item_key, toml_item = body_item
        item_type = get_item_type(toml_item=toml_item)
        parent_type = cast(ParentItem, container_info.item_type)
        key = item_key or ""
        return cls(
            item_type=item_type,
            parent_type=parent_type,
            key=key,
            hierarchy=hierarchy,
            from_aot=container_info.from_aot,
        )


class TOMLStatistics:
    """
    A class that counts and keeps track of the number of different TOML
    structures while recursively traversing a `tomlkit` object in the
    `_generate_descriptor` method of `_TOMLParser`.

    Maintains counts for the following `tomlkit` objects:
    - `tomlkit.items.Table`
    - `tomlkit.items.InlineTable`
    - `tomlkit.items.AoT`
    - `tomlkit.items.Comment`
    - `tomlkit.items.Item` (those that correspond to fields)
    - `tomlkit.items.Array`
    """

    def __init__(self) -> None:
        self.number_of_tables = 0
        self.number_of_inline_tables = 0
        self.number_of_aots = 0
        self.number_of_comments = 0
        self.number_of_fields = 0
        self.number_of_arrays = 0

    def add_table(self, table: items.Table) -> None:
        """
        Given a `tomlkit.items.Table` instance, will check to ensure that it
        is not a super table, and if so will update the table count. In addition,
        will check for a comment associated with the table and if there is one
        the comment count is updated.
        """
        if not (isinstance(table, items.Table) and table.is_super_table()):
            self.number_of_tables += 1
            self.add_comment(item=table)

    def add_inline_table(self, table: items.InlineTable) -> None:
        """
        Given a `tomlkit.items.InlineTable` instance, will update the inline table
        count. In addition, will check for a comment associated with the table and
        if there is one the comment count is updated.
        """
        self.number_of_inline_tables += 1
        self.add_comment(item=table)

    def add_array(self, item: items.Array) -> None:
        """
        Given a `tomlkit.items.Array` instance, will update the array count. In
        addition, will check for a comment associated with the array and if there
        is one the comment count is updated.
        """
        self.number_of_arrays += 1
        self.add_comment(item=item)

    def add_aot(self) -> None:
        """Updates the arrays-of-tables count."""
        self.number_of_aots += 1

    def add_comment(self, item: items.Item) -> None:
        """
        Given a generic `tomlkit.items.Item` instance, will check for a comment
        associated with the item and if there is one the comment count is updated.
        """
        if isinstance(item, items.Comment) or (
            not isinstance(item, items.Whitespace) and item.trivia.comment
        ):
            self.number_of_comments += 1

    def add_field(self, item: items.Item) -> None:
        """
        Given a generic `tomlkit.items.Item` instance, will update the field count. In
        addition, will check for a comment associated with the field and if there
        is one the comment count is updated.
        """
        self.number_of_fields += 1
        self.add_comment(item=item)



import copy
import datetime
import warnings
from abc import ABC, abstractmethod
from typing import Any, Optional, Tuple, Union, cast

import tomlkit
from tomlkit import TOMLDocument, items
from tomlkit.container import OutOfOrderTableProxy

from tomlkit_extras._constants import DICTIONARY_LIKE_TYPES
from tomlkit_extras._exceptions import KeyNotProvidedError, TOMLInsertionError
from tomlkit_extras._hierarchy import Hierarchy, standardize_hierarchy
from tomlkit_extras._typing import (
    BodyContainer,
    BodyContainerInOrder,
    BodyContainerItemDecomposed,
    BodyContainerItems,
    ContainerLike,
    Stylings,
    TOMLFieldSource,
    TOMLHierarchy,
)
from tomlkit_extras._utils import (
    complete_clear_array,
    complete_clear_out_of_order_table,
    complete_clear_tables,
    complete_clear_toml_document,
    convert_to_tomlkit_item,
    decompose_body_item,
    get_container_body,
)
from tomlkit_extras.toml._retrieval import get_attribute_from_toml_source

_VALID_ARRAY_OR_INLINE_TYPES = (
    bool,  # boolean
    int,  # integer
    float,  # float
    str,  # string
    datetime.datetime,  # offset_date_time / local_date_time
    datetime.date,  # local_date
    datetime.time,  # local_time
    items.Array,  # array
    items.InlineTable,  # inline_table
)


def container_insert(
    toml_source: TOMLFieldSource,
    insertion: Any,
    position: int,
    hierarchy: Optional[TOMLHierarchy] = None,
    key: Optional[str] = None,
) -> None:
    """
    Inserts an object that is tomlkit compatible based on a container
    position. A "container position" is an integer, indexed at 1,
    representing the position where the insertion object should be placed
    amongst all other types, including stylings (whitespace, comments),
    within a tomlkit type that supports insertion.

    The `toml_source` argument is the base tomlkit instance, and the
    `hierarchy` references the hierarchy relative to, and located within
    `toml_source` where the `insertion` argument will be placed.

    The `hierarchy` argument must exist within the `tomlkit` object,
    unless it is None. If this argument is passed in as None, then the
    insertion will occur at the top-level space of the `tomlkit` object.

    The `key` argument is the string key corresponding to the data that
    is being inserted. This should always be a string unless inserting in
    a array or array of tables. In those cases, a `key` should not be
    provided and will default to None.

    Args:
        toml_source (`TOMLFieldSource`): A `TOMLFieldSource` instance.
        insertion (Any): An instance of any type.
        position (int): The position of insertion, indexed at 1.
        hierarchy (`TOMLHierarchy` | None): A `TOMLHierarchy` instance or
            None. Defaults to None.
        key (str | None): A string corresponding to the key of data that
            is being inserted. Can also be None. Defaults to None.
    """
    _insert_into_toml_source(
        inserter=_PositionalInserter(
            toml_source=toml_source,
            hierarchy=hierarchy,
            key=key,
            insertion=insertion,
            position=position,
            by_attribute=False,
        )
    )


def attribute_insert(
    toml_source: TOMLFieldSource,
    insertion: Any,
    position: int,
    hierarchy: Optional[TOMLHierarchy] = None,
    key: Optional[str] = None,
) -> None:
    """
    Inserts an object that is tomlkit compatible based on a container
    position. A "container position" is an integer, indexed at 1,
    representing the position where the insertion object should be placed
    amongst all other types, including stylings (whitespace, comments),
    within a tomlkit type that supports insertion.

    The `toml_source` argument is the base tomlkit instance, and the
    `hierarchy` references the hierarchy relative to, and located within
    `toml_source` where the `insertion` argument will be placed.

    The `hierarchy` argument must exist within the `tomlkit` object,
    unless it is None. If this argument is passed in as None, then the
    insertion will occur at the top-level space of the `tomlkit` object.

    The `key` argument is the string key corresponding to the data that
    is being inserted. This should always be a string unless inserting in
    a array or array of tables. In those cases, a `key` should not be
    provided and will default to None.

    Args:
        toml_source (`TOMLFieldSource`): A `TOMLFieldSource` instance.
        insertion (Any): An instance of any type.
        position (int): The position of insertion, indexed at 1.
        hierarchy (`TOMLHierarchy` | None): A `TOMLHierarchy` instance or
            None. Defaults to None.
        key (str | None): A string corresponding to the key of data that
            is being inserted. Can also be None. Defaults to None.
    """
    _insert_into_toml_source(
        inserter=_PositionalInserter(
            toml_source=toml_source,
            hierarchy=hierarchy,
            key=key,
            insertion=insertion,
            position=position,
            by_attribute=True,
        )
    )


def general_insert(
    toml_source: TOMLFieldSource,
    insertion: Any,
    hierarchy: Optional[TOMLHierarchy] = None,
    key: Optional[str] = None,
) -> None:
    """
    Inserts an object that is tomlkit compatible based on a container
    position. A "container position" is an integer, indexed at 1,
    representing the position where the insertion object should be placed
    amongst all other types, including stylings (whitespace, comments),
    within a tomlkit type that supports insertion.

    The `toml_source` argument is the base tomlkit instance, and the
    `hierarchy` references the hierarchy relative to, and located within
    `toml_source` where the `insertion` argument will be placed.

    The `hierarchy` argument must exist within the `tomlkit` object,
    unless it is None. If this argument is passed in as None, then the
    insertion will occur at the top-level space of the `tomlkit` object.

    The `key` argument is the string key corresponding to the data that
    is being inserted. This should always be a string unless inserting in
    a array or array of tables. In those cases, a `key` should not be
    provided and will default to None.

    Args:
        toml_source (`TOMLFieldSource`): A `TOMLFieldSource` instance.
        insertion (Any): An instance of any type.
        hierarchy (`TOMLHierarchy` | None): A `TOMLHierarchy` instance or
            None. Defaults to None.
        key (str | None): A string corresponding to the key of data that
            is being inserted. Can also be None. Defaults to None.
    """
    _insert_into_toml_source(
        inserter=_GeneralInserter(
            toml_source=toml_source, hierarchy=hierarchy, key=key, insertion=insertion
        )
    )


class _BaseItemInserter(ABC):
    """
    A private base abstract class that is an abstract structure which provides
    tools to insert a `tomlkit.items.Item` object at a specific position within
    a `tomlkit` type that supports insertion.
    """

    def __init__(
        self, item_to_insert: BodyContainerItemDecomposed, by_attribute: bool
    ) -> None:
        self.item_inserted = False
        self.attribute, self.insertion = item_to_insert
        self.by_attribute = by_attribute

        self.attribute_position = self.container_position = 1

    @abstractmethod
    def add(self, item: items.Item, key: Optional[str] = None) -> None:
        """
        Abstract method to add a `tomlkit.items.Item` instance to a `tomlkit`
        structure.
        """
        pass

    def insert_attribute(self) -> None:
        """
        Inserts the relevant `items.Item` instance passed into the constructor
        to a `tomlkit` structure.
        """
        self.add(key=self.attribute, item=self.insertion)

    def insert_attribute_in_loop(self, position: int) -> None:
        """
        Runs within a for loop, and inserts the relevant `items.Item` instance
        passed into the constructor.

        Args:
            position (int): The current position when iterating through items in
                the body of a `BodyContainer` instance.
        """
        if (self.attribute_position == position and self.by_attribute) or (
            self.container_position == position and not self.by_attribute
        ):
            self.insert_attribute()
            self.item_inserted = True
            self.attribute_position += 1


class _DictLikeItemInserter(_BaseItemInserter):
    """
    A sub-class of `_BaseItemInserter` which provides tools to insert `tomlkit.items.Item`
    objects at specific positions within `tomlkit` dictionary-like types that support
    insertion.
    """

    def __init__(
        self,
        item_to_insert: Tuple[str, items.Item],
        container: Union[TOMLDocument, items.Table, items.InlineTable],
        by_attribute: bool = True,
    ) -> None:
        super().__init__(item_to_insert=item_to_insert, by_attribute=by_attribute)
        self.container = container

    def add(self, item: items.Item, key: Optional[str] = None) -> None:
        """
        Adds a `tomlkit.items.Item` instance to a dict-like `tomlkit` type.

        Args:
            item (`items.Item`): A `tomlkit.items.Item` instance.
            key (str | None): A string representation of a key in a TOML file,
                can be a string or None.
        """
        if key is None:
            self.container.add(cast(Stylings, item))
        else:
            self.container.add(key, item)


class _ListLikeItemInserter(_BaseItemInserter):
    """
    A sub-class of `_BaseItemInserter` which provides tools to insert `tomlkit.items.Item`
    objects at specific positions within `tomlkit` list-like types that support insertion.
    """

    def __init__(
        self,
        item_to_insert: BodyContainerItemDecomposed,
        container: items.Array,
        by_attribute: bool = True,
    ) -> None:
        super().__init__(item_to_insert=item_to_insert, by_attribute=by_attribute)
        self.container = container

    def add(self, item: items.Item, _: Optional[str] = None) -> None:
        """
        Adds a `tomlkit.items.Item` instance to a list-like `tomlkit` type.

        Args:
            item (`items.Item`): A `tomlkit.items.Item` instance.
        """
        self.container.append(item)


class _BaseInserter(ABC):
    """
    A private base abstract class that is an abstract structure which provides
    tools to run the entire insertion process to insert `tomlkit` types by
    attribute or container positions within `tomlkit` types that support insertion.
    """

    def __init__(
        self,
        toml_source: TOMLFieldSource,
        hierarchy: Optional[TOMLHierarchy],
        key: Optional[str],
        insertion: Any,
    ) -> None:
        self.toml_source = toml_source
        self.key = key
        self.hierarchy_obj: Optional[Hierarchy] = None

        self.toml_item: items.Item = convert_to_tomlkit_item(value=insertion)

        if hierarchy is not None:
            self.hierarchy_obj = standardize_hierarchy(hierarchy=hierarchy)

    @abstractmethod
    def array_of_tables_insert(
        self, array_of_tables: items.AoT, table: items.Table
    ) -> None:
        """
        An abstract method to insert an `tomlkit.items.Table` within an
        `tomlkit.items.AoT` instance.
        """
        pass

    @abstractmethod
    def insert(self, parent: BodyContainer) -> None:
        """
        An abstract method to insert an `tomlkit.items.Item` within a
        `tomlkit` type that is not a `tomlkit.items.AoT` instance.
        """
        pass

    def get_toml_source_insertion_object(self) -> ContainerLike:
        """
        Retrieve the point of insertion `tomlkit` type based on the `hierarchy`
        argument provided.

        There is logic to ensure that the `toml_source` is not nested within an
        array-of-tables, and that the parent is a valid `tomlkit` type that supports
        insertion.
        """
        parent: Union[TOMLFieldSource, items.Item]

        if self.hierarchy_obj is None:
            parent = self.toml_source
        else:
            retrieved_source = get_attribute_from_toml_source(
                hierarchy=self.hierarchy_obj, toml_source=self.toml_source
            )

            # Ensure the hierarchy does not map to a nested item within an array of
            # tables
            if isinstance(retrieved_source, list) and not isinstance(
                retrieved_source, items.Item
            ):
                raise TOMLInsertionError(
                    "Hierarchy maps to multiple items, insertion is not supported",
                    retrieved_source,
                )

            parent = retrieved_source

        # Ensure that the hierarchy does not map to a type that does not support
        # insertion
        if not isinstance(
            parent,
            (
                TOMLDocument,
                items.Table,
                items.InlineTable,
                OutOfOrderTableProxy,
                items.Array,
                items.AoT,
            ),
        ):
            raise TOMLInsertionError(
                "Hierarchy maps to a structure that does not support insertion",
                parent,
            )

        return parent


class _GeneralInserter(_BaseInserter):
    """
    A sub-class of `_BaseInserter` which provides tools to insert `tomlkit.items.Item`
    objects "generally", at the bottom of tomlkit types, that support insertion.
    """

    def array_of_tables_insert(
        self, array_of_tables: items.AoT, table: items.Table
    ) -> None:
        """
        Inserts an `tomlkit.items.Item` within an `tomlkit.items.AoT` instance
        for the "general" insert case.
        """
        array_of_tables.append(table)

    def insert(self, parent: BodyContainer) -> None:
        """
        Inserts an `tomlkit.items.Item` within a `tomlkit` type that is not
        a `tomlkit.items.AoT` instance.
        """
        if isinstance(parent, DICTIONARY_LIKE_TYPES):
            key = cast(str, self.key)
            parent[key] = self.toml_item
        else:
            parent.append(self.toml_item)


class _PositionalInserter(_BaseInserter):
    """
    A sub-class of `_BaseInserter` which provides tools to insert `tomlkit.items.Item`
    objects at specific positions within `tomlkit` types that support insertion.
    """

    def __init__(
        self,
        toml_source: TOMLFieldSource,
        hierarchy: Optional[TOMLHierarchy],
        key: Optional[str],
        insertion: Any,
        position: int,
        by_attribute: bool,
    ) -> None:
        super().__init__(
            toml_source=toml_source, hierarchy=hierarchy, key=key, insertion=insertion
        )
        self.position = position
        self.by_attribute = by_attribute

    @property
    def body_item(self) -> BodyContainerItemDecomposed:
        """
        Returns a tuple containing the string key and `tomlkit.items.Item`
        of the item to be inserted
        """
        return (self.key, self.toml_item)

    def array_of_tables_insert(
        self, array_of_tables: items.AoT, table: items.Table
    ) -> None:
        """
        Inserts an `tomlkit.items.Item` within an `tomlkit.items.AoT` instance
        for the "positional" insert case.
        """
        array_of_tables.insert(self.position - 1, table)

    def insert(self, parent: BodyContainer) -> None:
        """
        Inserts an `tomlkit.items.Item` within a `tomlkit` type that is not
        a `tomlkit.items.AoT` instance.
        """
        item_to_insert: BodyContainerItemDecomposed = self.body_item
        toml_body_items = copy.deepcopy(get_container_body(toml_source=parent))
        _refresh_container(initial_container=parent)

        container: BodyContainerInOrder

        # If the parent of the insertion point is an out-of-order table,
        # then overwrite the container variable by creating a temporary
        # new items.Table instance.
        if isinstance(parent, OutOfOrderTableProxy):
            container = tomlkit.table()
        else:
            container = parent

        inserter: _BaseItemInserter

        # Conditional to create insert object for dict-like or list-like tomlkit types
        if isinstance(container, (TOMLDocument, items.Table, items.InlineTable)):
            inserter = _DictLikeItemInserter(
                item_to_insert=cast(Tuple[str, items.Item], item_to_insert),
                container=container,
                by_attribute=self.by_attribute,
            )
        else:
            inserter = _ListLikeItemInserter(
                item_to_insert=item_to_insert,
                container=container,
                by_attribute=self.by_attribute,
            )

        # For out-of-order tables, update the original parent container (which
        # has been cleared of its contents)
        if isinstance(parent, OutOfOrderTableProxy):
            parent.update(container)

        _insert_item_at_position_in_container(
            position=self.position, inserter=inserter, toml_body_items=toml_body_items
        )


def _insert_item_at_position_in_container(
    position: int, inserter: _BaseItemInserter, toml_body_items: BodyContainerItems
) -> None:
    """
    A private function which executes the insertion logic to place a `tomlkit.items.Item`
    object into a `BodyContainer` instance at a specific position.
    """
    for toml_table_item in toml_body_items:
        item_key, toml_item = decompose_body_item(body_item=toml_table_item)

        if isinstance(toml_item, items.Whitespace):
            toml_item = tomlkit.ws(toml_item.value)

        inserter.insert_attribute_in_loop(position=position)

        if item_key is not None:
            inserter.attribute_position += 1

        inserter.add(key=item_key, item=toml_item)
        inserter.container_position += 1

    if not inserter.item_inserted:
        inserter.insert_attribute()


def _refresh_container(initial_container: BodyContainer) -> None:
    """
    A private function which "refreshes" or "clears" a `BodyContainer` instance
    of all data, as if starting from a clean slate.
    """
    if isinstance(initial_container, (items.Table, items.InlineTable)):
        complete_clear_tables(table=initial_container)
    elif isinstance(initial_container, OutOfOrderTableProxy):
        complete_clear_out_of_order_table(table=initial_container)
    elif isinstance(initial_container, items.Array):
        complete_clear_array(array=initial_container)
    elif isinstance(initial_container, TOMLDocument):
        complete_clear_toml_document(toml_document=initial_container)
    else:
        raise TypeError("Type is not a valid container-like structure")


def _insert_into_toml_source(inserter: _BaseInserter) -> None:
    """
    A private function which serves as the basis for all three insertion
    operations, `general_insert`, `attribute_insert`, and `container_insert`.
    """
    toml_source = inserter.get_toml_source_insertion_object()

    # For insertion into an array-of-tables, the item to be inserted must
    # only be an tomlkit.items.Table object
    if isinstance(toml_source, items.AoT):
        if not isinstance(inserter.toml_item, items.Table):
            raise TypeError(
                "Insertion at top level of an array of tables must be a table"
            )

        if inserter.key is not None:
            warnings.warn(
                "Not needed to pass 'key' when inserting into an array of tables",
                category=UserWarning,
            )

        name: Optional[str] = None

        # Extract a correct string name from the items.AoT object or the
        # items.Table object being inserted
        if toml_source.name is not None:
            name = toml_source.name
        elif inserter.toml_item.name is not None:
            name = inserter.toml_item.name

        if inserter.key is None and name is None:
            raise KeyNotProvidedError(
                "`key` is requred or a `name` must exist in the array or table."
            )

        if inserter.key != name and name is not None:
            inserter.key = name

        inserter.array_of_tables_insert(
            array_of_tables=toml_source, table=inserter.toml_item
        )

    # Otherwise the insertion is occuring into a dictionary-like object,
    # where the item to be inserted can be of any type, unless the source
    # is an inline table or array. If the source is an inline table, then
    # the type of the item should be one of the following:
    # - bool
    # - int
    # - float
    # - str
    # - datetime.datetime
    # - datetime.date
    # - datetime.time
    # - items.Array
    # - items.InlineTable
    else:
        if isinstance(toml_source, (items.Array, items.InlineTable)) and not isinstance(
            inserter.toml_item, _VALID_ARRAY_OR_INLINE_TYPES
        ):
            raise TypeError(
                "Cannot insert value of type "
                f"{type(inserter.toml_item).__name__} into an array."
            )

        # A check to ensure that if insertion is occuring into a dictionary-
        # like tomlkit object, then a string key must be provided
        if isinstance(toml_source, DICTIONARY_LIKE_TYPES) and inserter.key is None:
            raise KeyNotProvidedError(
                "`key` is required for dictionary-like tomlkit types"
            )

        inserter.insert(parent=toml_source)



from tomlkit import TOMLDocument, items
from tomlkit.container import OutOfOrderTableProxy

# Tomlkit types that are subclasses of dictionaries
DICTIONARY_LIKE_TYPES = (
    TOMLDocument,
    items.Table,
    items.InlineTable,
    OutOfOrderTableProxy,
)



from typing import Union, cast

from pyrsistent import PDeque, pdeque
from tomlkit import TOMLDocument, items

from tomlkit_extras._exceptions import InvalidHierarchyDeletionError
from tomlkit_extras._hierarchy import Hierarchy, standardize_hierarchy
from tomlkit_extras._typing import (
    TOMLDictLike,
    TOMLHierarchy,
    TOMLSource,
    TOMLValidReturn,
)


def _delete_attribute_from_aot(attribute: str, current_source: items.AoT) -> None:
    """
    A private function that deletes the deepest level of a specified hierarchy.
    """
    table_deleted = False

    for table_source in current_source[:]:
        if attribute in table_source:
            del table_source[attribute]
            table_deleted = True

        if not table_source:
            current_source.remove(table_source)

    if not table_deleted:
        raise InvalidHierarchyDeletionError(
            "Hierarchy does not exist in TOML source space"
        )


def _delete_iteration_for_aot(
    attribute: str, current_source: items.AoT, hierarchy_queue: PDeque[str]
) -> None:
    """
    A private function that executes the recursive deletion for a `tomlkit.items.AoT`
    instance.
    """
    for table_source in current_source:
        if attribute in table_source:
            next_source = table_source[attribute]
            _recursive_deletion(
                current_source=next_source, hierarchy_queue=hierarchy_queue
            )
            if not next_source:
                del table_source[attribute]


def _recursive_deletion(
    current_source: Union[TOMLDocument, TOMLValidReturn], hierarchy_queue: PDeque[str]
) -> None:
    """
    A private function that executes a recursive deletion of an item within a tomkit
    type, given a hierarchy marking the location of that item.
    """
    try:
        current_table: str = hierarchy_queue[0]
        hierarchy_queue_new: PDeque[str] = hierarchy_queue.popleft()

        if not hierarchy_queue_new:
            if isinstance(current_source, items.AoT):
                _delete_attribute_from_aot(
                    attribute=current_table, current_source=current_source
                )
            else:
                del cast(TOMLDictLike, current_source)[current_table]
        elif isinstance(current_source, items.AoT):
            _delete_iteration_for_aot(
                attribute=current_table,
                current_source=current_source,
                hierarchy_queue=hierarchy_queue_new,
            )
        else:
            next_source = cast(TOMLValidReturn, current_source[current_table])  # type: ignore[index]
            _recursive_deletion(
                current_source=next_source, hierarchy_queue=hierarchy_queue_new
            )
            if not next_source:
                del cast(TOMLDictLike, current_source)[current_table]
    except KeyError:
        raise InvalidHierarchyDeletionError(
            "Hierarchy does not exist in TOML source space"
        )


def delete_from_toml_source(hierarchy: TOMLHierarchy, toml_source: TOMLSource) -> None:
    """
    Deletes the tomlkit item residing at a speicifc hierarchy within a `TOMLSource`
    instance. In addition, the deletion will continue to cascade backwards as long
    as the last deletion resulted in an empty tomlkit structure.

    Accepts a `TOMLHierarchy` instance, being an instance of string or `Hierarchy`,
    and an instance of `TOMLSource`.

    Args:
        hierarchy (`TOMLHierarchy`): A `TOMLHierarchy` instance.
        toml_source (`TOMLSource`): A `TOMLSource` instance.
    """
    hierarchy_obj: Hierarchy = standardize_hierarchy(hierarchy=hierarchy)

    hierarchy_queue: PDeque[str] = pdeque(hierarchy_obj.full_hierarchy)
    _recursive_deletion(current_source=toml_source, hierarchy_queue=hierarchy_queue)



from collections import deque
from typing import (
    Any,
    Deque,
    Iterator,
    List,
    Literal,
    Tuple,
    Type,
    Union,
    cast,
    overload,
)

from tomlkit import TOMLDocument, items
from tomlkit.container import OutOfOrderTableProxy

from tomlkit_extras._constants import DICTIONARY_LIKE_TYPES
from tomlkit_extras._exceptions import (
    InvalidHierarchyRetrievalError,
    NotContainerLikeError,
)
from tomlkit_extras._hierarchy import Hierarchy, standardize_hierarchy
from tomlkit_extras._typing import (
    BodyContainerItem,
    Retrieval,
    TOMLFieldSource,
    TOMLHierarchy,
    TOMLSource,
)
from tomlkit_extras._utils import decompose_body_item, get_container_body
from tomlkit_extras.toml._out_of_order import fix_out_of_order_table


def _get_table_from_aot(
    current_source: List[items.Item], table: str
) -> List[items.Item]:
    """
    A private function that extracts all items within a `tomlkit.items.AoT`
    instance that correspond to a specific hierarchy.
    """
    next_source: List[items.Item] = []

    for source_item in current_source:
        if isinstance(source_item, items.AoT):
            next_source.extend(
                aot_item[table] for aot_item in source_item if table in aot_item
            )
        elif isinstance(source_item, DICTIONARY_LIKE_TYPES) and table in source_item:
            next_source.append(source_item[table])

    return next_source


def get_positions(hierarchy: TOMLHierarchy, toml_source: TOMLSource) -> Tuple[int, int]:
    """
    Returns both the attribute and container positions of an item located at
    a specific hierarchy within a `TOMLSource` instance. The attribute and
    container positions are relative to other items within the containing type
    of the item in question. The attribute position refers to the position of
    an item amongst all other key value pairs (fields, tables) within the
    containing object. The container position is the position of the item
    amongst all other types, including stylings (whitespace, comments), within
    the containing object.

    A tuple is returned, with the first item being an integer representing the
    attribute position, and the second item the container position.

    Accepts a `TOMLHierarchy` instance, being an instance of string or
    `Hierarchy`, and an instance of `TOMLSource`.

    Args:
        hierarchy (`TOMLHierarchy`): A `TOMLHierarchy` instance.
        toml_source (`TOMLSource`): A `TOMLSource` instance.

    Returns:
        Tuple[int, int]: A two-element tuple where both elements are of integer
            type. The first item representing the attribute position, and the
            second item the container position.
    """
    hierarchy_obj: Hierarchy = standardize_hierarchy(hierarchy=hierarchy)

    parent_source = find_parent_toml_source(
        hierarchy=hierarchy_obj, toml_source=toml_source
    )
    if not isinstance(
        parent_source,
        (
            TOMLDocument,
            items.Table,
            items.InlineTable,
            OutOfOrderTableProxy,
            items.Array,
        ),
    ):
        raise NotContainerLikeError("Hierarchy maps to a non-container-like object")

    table_body_items: Iterator[BodyContainerItem] = iter(
        get_container_body(toml_source=parent_source)
    )

    container_position = attribute_position = 0
    finding_positions = True

    try:
        while finding_positions:
            toml_table_item = next(table_body_items)
            item_key, _ = decompose_body_item(body_item=toml_table_item)

            if item_key is not None:
                attribute_position += 1

            container_position += 1

            if item_key == hierarchy_obj.attribute:
                finding_positions = False
    except StopIteration:
        raise InvalidHierarchyRetrievalError(
            "Hierarchy specified does not exist in TOMLDocument instance"
        )

    return attribute_position, container_position


@overload
def get_attribute_from_toml_source(
    hierarchy: TOMLHierarchy,
    toml_source: TOMLFieldSource,
    array: bool = True,
    fix_order: Literal[True] = True,
) -> Union[items.Item, List[items.Item]]: ...


@overload
def get_attribute_from_toml_source(
    hierarchy: TOMLHierarchy,
    toml_source: TOMLFieldSource,
    array: bool = True,
    fix_order: Literal[False] = False,
) -> Retrieval: ...


@overload
def get_attribute_from_toml_source(
    hierarchy: TOMLHierarchy,
    toml_source: TOMLFieldSource,
    array: bool = True,
    fix_order: bool = True,
) -> object: ...


def get_attribute_from_toml_source(
    hierarchy: TOMLHierarchy,
    toml_source: TOMLFieldSource,
    array: bool = True,
    fix_order: bool = False,
) -> Retrieval:
    """
    Retrieves and returns the `tomlkit` type located at a specific hierarchy
    within a `TOMLFieldSource` instance. If the hierarchy is nested within
    a `tomlkit.items.AoT` type, then a list with multiple `tomlkit` types can be
    returned. Otherwise a `tomlkit.container.OutOfOrderTableProxy` or a
    subclass of `tomlkit.items.Item` will be returned.

    If the hierarchy does not exist an `InvalidHierarchyError` will be raised.

    Args:
        hierarchy (`TOMLHierarchy`): A `TOMLHierarchy` instance.
        toml_source (`TOMLFieldSource`): A `TOMLFieldSource` instance.
        array (bool, optional): If set to False, when a `tomlkit.items.AoT`
            instance is to be returned, a list of the tables within the array
            are returned, otherwise it will be the AoT instance itself. Defaults
            to True.
        fix_order (bool, optional): If set to True, will fix any out-of-order
            tables before returning. Defaults to False.

    Returns:
        `Retrieval`: A `Retrieval` instance. Either a `tomlkit.continer.OutOfOrderTableProxy`,
            `tomlkit.items.Item` or list of `tomlkit.items.Item` instances.
    """
    hierarchy_obj: Hierarchy = standardize_hierarchy(hierarchy=hierarchy)

    hierarchy_of_tables: Deque[str] = deque(hierarchy_obj.full_hierarchy)
    current_source: Union[Retrieval, TOMLFieldSource] = toml_source

    try:
        while hierarchy_of_tables:
            table: str = hierarchy_of_tables.popleft()

            if isinstance(current_source, list):
                current_source = _get_table_from_aot(
                    current_source=current_source, table=table
                )
            else:
                current_source = cast(Retrieval, current_source[table])  # type: ignore[index]
    except KeyError:
        raise InvalidHierarchyRetrievalError(
            "Hierarchy specified does not exist in TOMLDocument instance"
        )
    else:
        if isinstance(current_source, list) and not current_source:
            raise InvalidHierarchyRetrievalError(
                "Hierarchy specified does not exist in TOMLDocument instance"
            )

        if isinstance(current_source, items.AoT) and not array:
            return [aot_table for aot_table in current_source]
        elif isinstance(current_source, OutOfOrderTableProxy) and fix_order:
            return fix_out_of_order_table(table=current_source)
        else:
            return cast(Retrieval, current_source)


def is_toml_instance(
    _type: Type[Any],
    *,
    hierarchy: TOMLHierarchy,
    toml_source: TOMLFieldSource,
    array: bool = False,
    fix_order: bool = False
) -> bool:
    """
    Checks if an item located at a specified hierarchy within a `TOMLFieldSource`
    instance is of a specific type.

    If the hierarchy is nested within a `tomlkit.items.AoT` type then multiple items
    will correspond to the hierarchy. In this case, each item is tested for type
    equality.

    Args:
        _type (Type[Any]): The type to check.
        hierarchy (`TOMLHierarchy`): A `TOMLHierarchy` instance.
        toml_source (`TOMLFieldSource`): A `TOMLFieldSource` instance.
        array (bool, optional): If set to False, when a `tomlkit.items.AoT` instance
            is to be tested, each table in the array is tested for type equality.
            Otherwise it will be the AoT instance itself that is checked. Defaults
            to False.
        fix_order (bool, optional): If set to True, it will fix any out-of-order
            tables before checking. Defaults to False.

    Returns:
        bool: A boolean indicating whether the item located at the hierarchy is of the
            specified type.
    """
    toml_items = get_attribute_from_toml_source(
        hierarchy=hierarchy, toml_source=toml_source, array=array, fix_order=fix_order
    )

    if not isinstance(toml_items, list) or (
        isinstance(toml_items, items.AoT) and array
    ):
        return isinstance(toml_items, _type)
    else:
        return all(isinstance(item, _type) for item in toml_items)


def find_parent_toml_source(
    hierarchy: Hierarchy, toml_source: TOMLFieldSource
) -> Union[Retrieval, TOMLFieldSource]:
    """
    A private function that finds the parent of an item located at a specific
    hierarchy within a `TOMLFieldSource` instance.
    """
    parent_toml: Union[Retrieval, TOMLFieldSource]
    if hierarchy.depth == 1:
        parent_toml = toml_source
    else:
        hierarchy_parent = Hierarchy.parent_hierarchy(hierarchy=str(hierarchy))
        parent_toml = get_attribute_from_toml_source(
            hierarchy=hierarchy_parent, toml_source=toml_source
        )

    return parent_toml



from __future__ import annotations

import copy
import inspect
import itertools
import re
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any, ClassVar, Dict, List, Optional, Set, Type, TypeVar, cast

from tomlkit import items

from tomlkit_extras._exceptions import InvalidStylingError
from tomlkit_extras._hierarchy import Hierarchy
from tomlkit_extras._typing import (
    AoTItem,
    FieldItem,
    Item,
    ParentItem,
    StyleItem,
    Stylings,
    Table,
    TableItem,
)
from tomlkit_extras._utils import find_comment_line_no, safe_unwrap
from tomlkit_extras.descriptor._helpers import (
    CommentDescriptor,
    create_comment_descriptor,
)
from tomlkit_extras.descriptor._types import ItemInfo

_WHITESPACE_PATTERN = r"^[ \n\r]*$"

Descriptor = TypeVar("Descriptor", bound="AbstractDescriptor")


def _chain_stylings(styles: Dict[str, List[StyleDescriptor]]) -> List[StyleDescriptor]:
    """A private function which flattens a list of lists of `StyleDescriptor` objects."""
    return list(itertools.chain.from_iterable(styles.values()))


def _from_info_to_hierarchy(info: ItemInfo) -> Hierarchy:
    """
    A private function which converts a string hierarchy from an `ItemInfo` object
    into a `Hierarchy`.
    """
    hierarchy = Hierarchy.from_str_hierarchy(info.hierarchy)
    hierarchy.add_to_hierarchy(update=info.key)
    return hierarchy


@dataclass
class AoTDescriptors:
    """
    Stores all array-of-tables that have been parsed while recursively
    traversing a TOML structure in the `_generate_descriptor` method of
    `_TOMLParser`.
    """

    aots: List[AoTDescriptor]

    def get_array(self) -> AoTDescriptor:
        """Retrieves the most recent `AoTDescriptor` object added."""
        return self.aots[-1]

    def update_arrays(self, array: AoTDescriptor) -> None:
        """Adds a new `AoTDescriptor` object to existing store."""
        self.aots.append(array)


@dataclass
class StylingDescriptors:
    """
    Provides access to all stylings that existing within a TOML structure.
    These are split up into two main categories, comments and whitespaces.

    Each group is contained in a dictionary where the keys are the string
    representations of the stylings, and the values are the list of
    `StyleDescriptor` objects that correspond to stylings that have that
    specific string.

    The values are lists as there can be multiple whitespaces or comments that
    have the same value.

    Attributes:
        comments (Dict[str, List[`StyleDescriptor`]]): A dictionary where the
            keys are string representations of the comments, and the values are
            lists of `StyleDescriptor` corresponding to the comments.
        whitespace (Dict[str, List[`StyleDescriptor`]]): A dictionary where the
            keys are string representations of the whitespaces, and the values
            are lists of `StyleDescriptor` corresponding to the whitespaces.
    """

    comments: Dict[str, List[StyleDescriptor]]
    whitespace: Dict[str, List[StyleDescriptor]]

    @property
    def _decomposed_comments(self) -> List[StyleDescriptor]:
        """Private property that returns a list of the comments store values."""
        return _chain_stylings(styles=self.comments)

    @property
    def _decomposed_whitespace(self) -> List[StyleDescriptor]:
        """Private property that returns a list of the whitespace store values."""
        return _chain_stylings(styles=self.whitespace)

    def get_styling(self, styling: str) -> List[StyleDescriptor]:
        """
        Retrieves all stylings, represented by `StyleDescriptor` objects, that
        correspond to a specific string representation.

        Args:
            styling (str): A string representation of a comment or whitespace.

        Returns:
            List[`StyleDescriptor`]: A list of `StyleDescriptor` instances.
        """
        # Check if the string representation of the styling matches the
        # format of a whitespace or text comment
        is_comment = not re.match(_WHITESPACE_PATTERN, styling)

        styling_space = self.comments if is_comment else self.whitespace
        if styling not in styling_space:
            raise InvalidStylingError(
                "Styling does not exist in set of valid stylings",
                set(styling_space.keys()),
            )

        return styling_space[styling]

    def get_stylings(
        self, styling: Optional[StyleItem] = None
    ) -> List[StyleDescriptor]:
        """
        Retrieves all stylings, represented by `StyleDescriptor` objects, that
        correspond to a specific string literal.

        If "whitespace" is passed all whitespace stylings will be returned. If
        "comment" is passed all comment stylings will be returned. If it is None,
        then all stylings will be returned.

        Args:
            styling (`StyleItem` | None): A literal that identifies the type of
                styling to retrieve. Can be either "whitespace" or "comment".
                Defaults to None.

        Returns:
            List[`StyleDescriptor`]: A list of `StyleDescriptor` instances.
        """
        stylings: List[StyleDescriptor] = []

        if styling != "comment":
            stylings.extend(self._decomposed_whitespace)

        if styling != "whitespace":
            stylings.extend(self._decomposed_comments)

        return stylings

    def _update_stylings(self, style: Stylings, info: ItemInfo, line_no: int) -> None:
        """
        Private method that will create a new `StyleDescriptor` and update the
        existing store of stylings (comments and whitespaces) already parsed.
        """
        styling_value: str  # Must always be a string value

        # Based on whether the styling is a comment or a whitespace, a different
        # attribute from that tomlkit type will be assigned to be the value
        # of the styling
        if isinstance(style, items.Comment):
            styling_value = style.trivia.comment
            current_source = self.comments
        else:
            styling_value = style.value
            current_source = self.whitespace

        # Generate a StyleDescriptor object that will be added to the already
        # collected store of stylings
        styling_position = StyleDescriptor(
            style=styling_value, line_no=line_no, info=info
        )
        if styling_value not in current_source:
            current_source[styling_value] = [styling_position]
        else:
            current_source[styling_value].append(styling_position)


class AbstractDescriptor(ABC):
    __special__: ClassVar[Set[str]] = set()

    """
    Base descriptor class, which provides no functionality, but a series
    of common attributes for all sub-classes, those being `TableDescriptor`,
    `AoTDescriptor`, `StyleDescriptor`, and `FieldDescriptor`.
    
    Properties:
        parent_type (`ParentItem` | None): A `ParentItem` instance, corresponding
            to a string literal representing the type of the parent of the
            structure. Can be None if there is no parent.
        hierarchy (`Hierarchy`): A `Hierarchy` instance representing the full
            hierarchy of the structure.
        container_position (int): An integer position of the structure amongst all
            other types, including stylings (whitespace, comments), within the
            parent.
        from_aot (bool): A boolean indicating whether the structure is nested
            within an array of tables.
    """

    def __init__(self, item_info: ItemInfo) -> None:
        self._item_info = copy.deepcopy(item_info)

    def copy(self: Descriptor) -> Descriptor:
        """Returns a shallow copy of the object."""
        return copy.copy(self)

    def deepcopy(self: Descriptor) -> Descriptor:
        """Returns a deep copy of the object."""
        return copy.deepcopy(self)

    def __repr__(self) -> str:
        repr_body: List[str] = []
        all_members: Dict[str, Any] = dict(
            dict(inspect.getmembers(self.__class__)), **self.__dict__
        )

        for member_name, member_value in all_members.items():
            if not member_name.startswith("_") and member_name not in self.__special__:
                if isinstance(member_value, property):
                    member_value = member_value.__get__(self)

                if not callable(member_value):
                    repr_body.append(f"    {member_name}={member_value},\n")

        repr_string = f"{self.__class__.__name__}(\n{''.join(repr_body)})"
        return repr_string

    @property
    @abstractmethod
    def item_type(self) -> Item:
        """Returns the type of the structure value."""
        pass

    @property
    @abstractmethod
    def hierarchy(self) -> Optional[Hierarchy]:
        """Returns the hierarchy of the TOML structure as a `Hierarchy` object."""
        pass

    @property
    def from_aot(self) -> bool:
        """
        Returns a boolean indicating whether the structure is located within an
        array of tables.
        """
        return self._item_info.from_aot

    @property
    def container_position(self) -> int:
        """
        Returns the position, indexed at 1, of the attribute among other attributes,
        including stylings within its parent.

        Attributes in this case are fields, tables, or arrays.
        """
        return self._item_info.position.container

    @property
    def parent_type(self) -> Optional[ParentItem]:
        """
        Returns the type of the parent structure. Can be None if there was no parent.
        """
        return self._item_info.parent_type


class AttributeDescriptor(AbstractDescriptor):
    """
    An extension of the `AbstractDescriptor` which is built for structures who
    ar similear to that of an "attribute". These are fields, tables, or array of
    tables.

    Properties:
        name (str): The name of the attribute (field, table, or array of tables).
        attribute_position (int): An integer position of the structure amongst all
            other key value pairs (fields, tables) within the parent.
    """

    @property
    def name(self) -> str:
        """Returns the name of the TOML structure."""
        return self._item_info.key

    @property
    def attribute_position(self) -> int:
        """
        Returns the position, indexed at 1, of the attribute among other attributes
        within its parent.

        Attributes in this case are fields, tables, or arrays.
        """
        return self._item_info.position.attribute

    @property
    def hierarchy(self) -> Hierarchy:
        """Returns the hierarchy of the TOML structure as a `Hierarchy` object."""
        return _from_info_to_hierarchy(info=self._item_info)


class FieldDescriptor(AttributeDescriptor):
    __special__: ClassVar[Set[str]] = {"stylings"}

    """
    A class which provides detail on a field, a key-value pair that cannot
    contain nested key-value pairs.

    Inherits properties from `AttributeDescriptor`:
    - `parent_type`
    - `name`
    - `hierarchy`
    - `attribute_position`
    - `container_position`
    - `from_aot`

    Attributes:
        item_type (`FieldItem`): A `FieldItem` instance, corresponding to a
            string literal representing the type of the table, being
            either 'field' or 'array'.
        line_no (int): An integer line number marking the beginning of the
            structure.
        value (Any): The value of the field.
        value_type (Type[Any]): The type of the field value.
        comment (`CommentDescriptor` | None): A `CommentDescriptor` instance,
            correspondng to the comment associated with the structure. Can
            be None if there is no comment.
        stylings (`StylingDescriptors`): An object with all stylings associated
            with the field.
    """

    def __init__(
        self,
        line_no: int,
        info: ItemInfo,
        value: Any,
        comment: Optional[CommentDescriptor],
        stylings: StylingDescriptors,
    ) -> None:
        super().__init__(item_info=info)
        self.line_no = line_no
        self.value = value
        self.comment = comment
        self.stylings = stylings

    @property
    def item_type(self) -> FieldItem:
        """
        Returns a literal identifying the type of the field as an "array" or
        "field" (non-array).
        """
        return cast(FieldItem, self._item_info.item_type)

    @property
    def value_type(self) -> Type[Any]:
        """Returns the type of the field value."""
        return type(self.value)

    @classmethod
    def _from_toml_item(
        cls, item: items.Item, info: ItemInfo, line_no: int
    ) -> FieldDescriptor:
        """
        Private class method which generates an instance of `FieldDescriptor` for
        a given field while recursively traversing a TOML structure in the
        `_generate_descriptor` method of `_TOMLParser`.
        """
        comment_line_no: Optional[int]
        stylings = StylingDescriptors(comments=dict(), whitespace=dict())
        if isinstance(item, items.Array):
            comment_line_no = None
        else:
            comment_line_no = find_comment_line_no(line_no=line_no, item=item)

        comment = create_comment_descriptor(item=item, line_no=comment_line_no)
        value = safe_unwrap(structure=item)
        return cls(
            line_no=line_no, info=info, value=value, comment=comment, stylings=stylings
        )

    def _update_comment(self, item: items.Array, line_no: int) -> None:
        """
        A private method that updates a comment attributed to an array field.
        """
        comment_line_no = find_comment_line_no(line_no=line_no, item=item)
        self.comment = create_comment_descriptor(item=item, line_no=comment_line_no)


class TableDescriptor(AttributeDescriptor):
    __special__: ClassVar[Set[str]] = {"stylings", "fields"}

    """
    A dataclass which provides detail on a table or inline table, a key-value
    pair that can contain nested key-value pairs.

    Inherits properties from `AttributeDescriptor`:
    - `parent_type`
    - `name`
    - `hierarchy`
    - `attribute_position`
    - `container_position`
    - `from_aot`

    Attributes:
        item_type (`TableItem`): A `TableItem` instance, corresponding to a
            string literal representing the type of the table, being
            either 'table' or 'inline-table'.
        line_no (int): An integer line number marking the beginning of the
            table.
        fields (Dict[str, `FieldDescriptor`]): A dictionary which has
            key-value pairs each being a field contained in the table. The
            keys are strings representing names of fields (not tables) and
            the corresponding values are `FieldDescriptor` instances.
        num_fields (int): The number of fields contained in table.
        comment (`CommentDescriptor` | None): A `CommentDescriptor` instance,
            correspondng to the comment associated with the structure. Can
            be None if there is no comment.
        stylings (`StylingDescriptors`): An object with all stylings appearing
            within the table.
    """

    def __init__(
        self,
        line_no: int,
        info: ItemInfo,
        comment: Optional[CommentDescriptor],
        stylings: StylingDescriptors,
    ) -> None:
        super().__init__(item_info=info)
        self.line_no = line_no
        self.comment = comment
        self.stylings = stylings

        self._fields: Dict[str, FieldDescriptor] = dict()

    @property
    def fields(self) -> Dict[str, FieldDescriptor]:
        """
        Returns a dictionary containing all fields appearing in the table.

        The keys of the dictionary are the string field names, and the values
        are `FieldDescriptor` objects.
        """
        return self._fields

    @property
    def num_fields(self) -> int:
        """Returns the number of fields contained within the table."""
        return len(self._fields)

    @property
    def item_type(self) -> TableItem:
        """
        Returns a literal identifying the type of the table as a "table" or
        "inline-table".
        """
        return cast(TableItem, self._item_info.item_type)

    @classmethod
    def _from_table_item(
        cls, table: Table, info: ItemInfo, line_no: int
    ) -> TableDescriptor:
        """
        Private class method which generates an instance of `TableDescriptor` for
        a given table while recursively traversing a TOML structure in the
        `_generate_descriptor` method of `_TOMLParser`.
        """
        comment_line_no = find_comment_line_no(line_no=line_no, item=table)
        stylings = StylingDescriptors(comments=dict(), whitespace=dict())
        comment = create_comment_descriptor(item=table, line_no=comment_line_no)
        return cls(line_no=line_no, info=info, comment=comment, stylings=stylings)

    def _add_field(
        self, item: items.Item, info: ItemInfo, line_no: int
    ) -> FieldDescriptor:
        """A private method that adds a field to the existing store of fields."""
        field_descriptor = FieldDescriptor._from_toml_item(
            item=item, info=info, line_no=line_no
        )
        self._fields.update({info.key: field_descriptor})
        return field_descriptor


class StyleDescriptor(AbstractDescriptor):
    """
    A dataclass which provides detail on a specific styling appearing in a
    tomlkit type instance.

    A styling can either be a comment, represented in tomlkit as a
    `tomlkit.items.Comment` instance, or a whitespace, represented as a
    `tomlkit.items.Whitespace` instance.

    These are comments or whitespaces that are not directly associated with
    a field or table, but are contained within tomlkit structures like tables.

    Inherits properties from `AbstractDescriptor`:
    - `parent_type`
    - `container_position`
    - `from_aot`

    Attributes:
        item_type (`StyleItem`): A `StyleItem` instance, corresponding to a
            string literal representing the type of the styling, being
            either 'whitespace' or 'comment'.
        hierarchy (`Hierarchy` | None): A `Hierarchy` instance representing the full
            hierarchy of the structure, or None if it is a top-level styling.
        style (str): The string value of the style.
        line_no (int): An integer line number marking the beginning of the
            styling.
    """

    def __init__(self, style: str, line_no: int, info: ItemInfo) -> None:
        super().__init__(item_info=info)
        self.style = style
        self.line_no = line_no

    @property
    def hierarchy(self) -> Optional[Hierarchy]:
        """Returns the hierarchy of the TOML sstructure as a `Hierarchy` object."""
        if not self._item_info.hierarchy:
            return None
        else:
            return _from_info_to_hierarchy(info=self._item_info)

    @property
    def item_type(self) -> StyleItem:
        """
        Returns a literal identifying the type of the styling as a "comment" or
        "whitespace".
        """
        return cast(StyleItem, self._item_info.item_type)


class AoTDescriptor(AttributeDescriptor):
    __special__: ClassVar[Set[str]] = {"tables"}

    """
    A dataclass which provides detail on an array of tables, a list of
    tables.

    Inherits properties from `AttributeDescriptor`:
    - `parent_type`
    - `name`
    - `hierarchy`
    - `attribute_position`
    - `container_position`
    - `from_aot`

    Attributes:
        item_type (`AoTItem`): A `AoTItem` instance, corresponding to a
            string literal representing the structure type.
        line_no (int): An integer line number marking the beginning of the
            array of tables.
        tables (List[`TableDescriptor`]): A list of `TableDescriptor`
            instances where each one represents a table within the array
            of tables.
    """

    def __init__(self, line_no: int, info: ItemInfo) -> None:
        super().__init__(item_info=info)
        self.line_no = line_no

        self._tables: Dict[str, List[TableDescriptor]] = dict()

    @property
    def item_type(self) -> AoTItem:
        """Returns a literal being "array-of-tables"."""
        return cast(AoTItem, self._item_info.item_type)

    @property
    def tables(self) -> Dict[str, List[TableDescriptor]]:
        """
        Returns a dictionary containing all tables appearing in the array.

        The keys of the dictionary are the string table hierarchies, and the values
        are lists of `TableDescriptor` objects. The values are lists, as within an
        array-of-tables, there can be multiple tables associated with the same
        hierarchy.
        """
        return self._tables

    def num_tables(self, hierarchy: Optional[str] = None) -> int:
        """
        Retrieves the number of tables within the array that is associated
        with a specific hierarchy.

        If no hierarchy is provided, then defaults to None. In this case,
        all tables will be counted and result returned.

        Args:
            hierarchy (str | None): A string representation of a TOML hierarchy.
                Defaults to None.

        Returns:
            int: An integer representing the number of tables.
        """
        tables: Optional[List[TableDescriptor]]

        if hierarchy is None:
            num_tables = 0
            for tables in self._tables.values():
                num_tables += len(tables)
            return num_tables
        else:
            tables = self.tables.get(hierarchy, None)
            return len(tables) if tables is not None else 0

    def _get_table(self, hierarchy: str) -> TableDescriptor:
        """
        A private method that retrieves a specific `TableDescriptor` object,
        representing a table in an array-of-tables, given a string hierarchy.
        """
        return self._tables[hierarchy][-1]

    def _update_tables(self, hierarchy: str, table_descriptor: TableDescriptor) -> None:
        """A private method that adds a table to the existing store of tables."""
        if hierarchy not in self._tables:
            self._tables[hierarchy] = [table_descriptor]
        else:
            self._tables[hierarchy].append(table_descriptor)



from abc import ABC, abstractmethod
from typing import Any, Dict, Set, cast

from tomlkit import items

from tomlkit_extras._typing import Item, Stylings, Table, TOMLHierarchy
from tomlkit_extras._utils import standardize_hierarchy
from tomlkit_extras.descriptor._descriptors import (
    AoTDescriptor,
    AoTDescriptors,
    FieldDescriptor,
    StylingDescriptors,
    TableDescriptor,
)
from tomlkit_extras.descriptor._helpers import LineCounter, item_is_table
from tomlkit_extras.descriptor._types import ItemInfo


class BaseStore(ABC):
    """
    A base abstract class providing a general structure for all TOML stores
    used by the `TOMLDocumentDescriptor` class. These include the
    `DocumentStore`, `ArrayOfTablesStore` and `TableStore`.
    """

    def __init__(self) -> None:
        # For faster access, the most recent FieldDescriptor created is
        # stored
        self._field_descriptor: FieldDescriptor

    @property
    def field_descriptor(self) -> FieldDescriptor:
        """Returns the most recent `FieldDescriptor` created and added."""
        return self._field_descriptor

    @field_descriptor.setter
    def field_descriptor(self, field: FieldDescriptor) -> None:
        """Sets the `field_descriptor` property."""
        self._field_descriptor = field

    @abstractmethod
    def get(self, hierarchy: str) -> Any:
        pass

    @abstractmethod
    def get_stylings(self, style_info: ItemInfo) -> StylingDescriptors:
        pass

    @abstractmethod
    def contains(self, hierarchy: str) -> bool:
        pass

    @abstractmethod
    def update(self, item: items.Item, info: ItemInfo) -> None:
        pass


class BaseTableStore(BaseStore):
    """
    A base abstract class providing a additional structure for all TOML stores
    that store data included in tables. These stores include
    `ArrayOfTablesStore` and `TableStore`.
    """

    @property
    @abstractmethod
    def hierarchies(self) -> Set[str]:
        pass

    @abstractmethod
    def add_table(self, hierarchy: str, table: Table, info: ItemInfo) -> None:
        pass


class DocumentStore(BaseStore):
    """
    A sub-class of `BaseStore` which stores any fields or stylings that are part
    of the top-level of a `tomlkit.TOMLDocument` instance. Additional basic
    functionality is included to retrieve information from the store.
    """

    def __init__(self, line_counter: LineCounter) -> None:
        self._line_counter = line_counter
        self._document_fields: Dict[str, FieldDescriptor] = dict()
        self._document_stylings: StylingDescriptors = StylingDescriptors(
            comments=dict(), whitespace=dict()
        )

    @property
    def fields(self) -> Set[str]:
        """
        Returns a set of strings corresponding to all fields that have
        been processed.
        """
        return set(self._document_fields.keys())

    def get(self, hierarchy: str) -> FieldDescriptor:
        """
        Given an field name will return a `FieldDescriptor` instance corresponding
        to that field. These are fields that appear at the top-level of the document,
        thus they are not nested with larger structures like tables.

        Args:
            hierarchy (str): A TOML hierarchy corresponding to field.

        Returns:
            `FieldDescriptor`: A `FieldDescriptor` instance.
        """
        return self._document_fields[hierarchy]

    def contains(self, hierarchy: str) -> bool:
        """
        A boolean method that checks to see if field exists in all those processed.
        Returns True if the field exists False if it does not exists.

        Args:
            hierarchy (str): A TOML hierarchy corresponding to the name of a field.

        Returns:
            bool: A boolean indicating whether the field exists.
        """
        return hierarchy in self._document_fields

    def update(self, item: items.Item, info: ItemInfo) -> None:
        """
        Adds a new field to the store of fields already processed. A new key-value
        pair is added to the dictionary, where the key is the string name of the field,
        and the value is a `FieldDescriptor` instance.

        Args:
            item (`tomlkit.items.Item`): A `tomlkit.items.Item` instance corresponding
                to a field appearing in a TOML file.
            info (`ItemInfo`): An `ItemInfo` instance with basic info on the field.
        """
        field_descriptor: FieldDescriptor = FieldDescriptor._from_toml_item(
            item=item, info=info, line_no=self._line_counter.line_no
        )
        self.field_descriptor = field_descriptor
        self._document_fields[info.key] = field_descriptor

    def get_stylings(self, style_info: ItemInfo) -> StylingDescriptors:
        """
        Returns all stylings associated with a specific field. Stylings are separated
        by whether it is a whitespace or comment.

        Args:
            style_info (`ItemInfo`): An `ItemInfo` instance with basic info on the styling.

        Returns:
            `StylingDescriptors`: A `StylingDescriptors` instance containing all
                stylings associated with a field.
        """
        if style_info.parent_type == "document":
            styling_positions = self._document_stylings
        else:
            styling_positions = self.field_descriptor.stylings
        return styling_positions


class ArrayOfTablesStore(BaseTableStore):
    """
    A sub-class of `BaseTableStore` which stores any fields, tables, or stylings
    that are included in arrays of tables appearing in a `tomlkit.TOMLDocument`
    instance. Additional basic functionality is included to retrieve information
    from the store.
    """

    def __init__(self, line_counter: LineCounter) -> None:
        self._line_counter = line_counter
        self._array_of_tables: Dict[str, AoTDescriptors] = dict()

        # For faster access, a second dictionary ise used to store the
        # most recent TableDescriptor object created for a given hierarchy
        self._table_checkpoints: Dict[str, TableDescriptor] = dict()

    @property
    def hierarchies(self) -> Set[str]:
        """
        Returns a set of hierarchies corresponding to all array of tables that have
        been processed.
        """
        return set(self._array_of_tables.keys())

    def get(self, hierarchy: str) -> AoTDescriptors:
        """
        Given an array of tables hierarchy will return a `AoTDescriptors`
        instance corresponding to that array.

        Args:
            hierarchy (str): A TOML hierarchy corresponding to an array of tables.

        Returns:
            `AoTDescriptors`: An `AoTDescriptors` instance.
        """
        return self._array_of_tables[hierarchy]

    def contains(self, hierarchy: str) -> bool:
        """
        A boolean method that checks to see if an array of tables hierarchy exists
        in all those processed. Returns True if the hierarchy exists False if it
        does not exists.

        Args:
            hierarchy (str): A TOML hierarchy corresponding to an array of tables.

        Returns:
            bool: A boolean indicating whether the hierarchy exists.
        """
        return hierarchy in self.hierarchies

    def append(self, hierarchy: str, array_of_tables: AoTDescriptor) -> None:
        """
        Updates the existing array of tables store based on a hierarchy.

        If the hierarchy has not already been encountered then a new key-value pair
        is created where the key is the string hierarchy and the value is an
        `AoTDescriptors` instance.

        If the hierarchy has already been encountered, then the existing value
        corresponding to a `AoTDescriptors` instance is updated.

        Args:
            hierarchy (str): A TOML hierarchy corresponding to an array of tables.
            array_of_tables (`AoTDescriptors`): An `AoTDescriptors` instance.
        """
        if not hierarchy in self._array_of_tables:
            self._array_of_tables[hierarchy] = AoTDescriptors(aots=[array_of_tables])
        else:
            self._array_of_tables[hierarchy].update_arrays(array=array_of_tables)

    def _get_array_hierarchy(self, hierarchy: TOMLHierarchy) -> str:
        """
        Given a `TOMLHierarchy` instance representing a TOML hierarchy, will return
        the longest ancestor hierarchy that exists in the store. This effectively
        finds the newest array of tables ancestor of the given hierarchy.

        Args:
            hierarchy (`TOMLHierarchy`): A `TOMLHierarchy` instance.

        Returns:
            str: Returns a string hierarchy where a match was found.
        """
        hierarchy_obj = standardize_hierarchy(hierarchy=hierarchy)
        return cast(
            str, hierarchy_obj.longest_ancestor_hierarchy(hierarchies=self.hierarchies)
        )

    def _get_aot_table(self, hierarchy: str) -> TableDescriptor:
        """
        Retrieves a table located in an array of tables. Given a string table hierarchy
        will return a `TableDescriptor` instance.

        Args:
            hierarchy (str): A TOML hierarchy corresponding to a table.

        Returns:
            `TableDescriptor`: A `TableDescriptor` instance.
        """
        return self._table_checkpoints[hierarchy]

    def update(self, item: items.Item, info: ItemInfo) -> None:
        """
        Adds a new field to the store of fields already processed. A new key-value
        pair is added to the dictionary, where the key is the string name of the field,
        and the value is a `FieldDescriptor` instance.

        Args:
            item (`tomlkit.items.Item`): A `tomlkit.items.Item` instance corresponding
                to a field appearing in a TOML file.
            info (`ItemInfo`): An `ItemInfo` instance with basic info on the field.
        """
        table: TableDescriptor = self._get_aot_table(hierarchy=info.hierarchy)
        field_desctiptor: FieldDescriptor = table._add_field(
            item=item, info=info, line_no=self._line_counter.line_no
        )
        self.field_descriptor = field_desctiptor

    def add_table(self, hierarchy: str, table: Table, info: ItemInfo) -> None:
        """
        Adds a new table to the store of tables already processed. A new key-value
        pair is added to the dictionary, where the key is the string hierarchy of the
        table, and the value is a `TableDescriptor` instance.

        Args:
            hierarchy (str): A TOML hierarchy corresponding to a table.
            table (`Table`): A Table instance, being either a tomlkit.items.Table or
                tomlkit.items.InlineTable.
            info (`ItemInfo`): An `ItemInfo` instance with basic info on the table.
        """
        array_hierarchy = self._get_array_hierarchy(hierarchy=hierarchy)
        array_of_tables = self._array_of_tables[array_hierarchy]

        table_descriptor = TableDescriptor._from_table_item(
            table=table, info=info, line_no=self._line_counter.line_no
        )

        # Immediately after creation of the TableDescriptor, update the
        # table checkpoint store
        self._table_checkpoints.update({hierarchy: table_descriptor})

        # Retrieve the relevant AoTDescriptor object and add the new table
        array = array_of_tables.get_array()
        array._update_tables(hierarchy=hierarchy, table_descriptor=table_descriptor)

    def get_stylings(self, style_info: ItemInfo) -> StylingDescriptors:
        """
        Returns all stylings associated with a specific table. Stylings are separated
        by whether its whitespace or comments.

        Args:
            style_info (`ItemInfo`): An `ItemInfo` instance with basic info on the styling.

        Returns:
            `StylingDescriptors`: A `StylingDescriptors` instance containing all
                stylings associated with a table.
        """
        if item_is_table(item_type=style_info.parent_type):
            table: TableDescriptor = self._get_aot_table(hierarchy=style_info.hierarchy)
            styling_descriptors = table.stylings
        else:
            styling_descriptors = self.field_descriptor.stylings
        return styling_descriptors


class TableStore(BaseTableStore):
    """
    A sub-class of `BaseTableStore` which stores any fields or stylings that
    are included in tables appearing in a tomlkit.TOMLDocument instance, but
    not appearing within an array of tables. Additional basic functionality
    is included to retrieve information from the store.
    """

    def __init__(self, line_counter: LineCounter) -> None:
        self._line_counter = line_counter
        self._tables: Dict[str, TableDescriptor] = dict()

    @property
    def hierarchies(self) -> Set[str]:
        """
        Returns a set of hierarchies corresponding to all tables that do not appear
        in any array of tables structures.
        """
        return set(self._tables.keys())

    def get(self, hierarchy: str) -> TableDescriptor:
        """
        Given a table hierarchy will return a `TableDescriptor` instance
        corresponding to that table.

        Args:
            hierarchy (str): A TOML hierarchy corresponding to a table.

        Returns:
            `TableDescriptor`: A `TableDescriptor` instance.
        """
        return self._tables[hierarchy]

    def contains(self, hierarchy: str) -> bool:
        """
        A boolean method that checks to see if a table hierarchy exists in
        all tables processed. Returns True if the hierarchy exists False if it
        does not exists.

        Args:
            hierarchy (str): A TOML hierarchy corresponding to a table.

        Returns:
            bool: A boolean indicating whether the hierarchy exists.
        """
        return hierarchy in self._tables

    def update(self, item: items.Item, info: ItemInfo) -> None:
        """
        Adds a new field to the store of fields already processed. A new key-value
        pair is added to the dictionary, where the key is the string name of the field,
        and the value is a `FieldDescriptor` instance.

        Args:
            item (`tomlkit.items.Item`): A `tomlkit.items.Item` instance corresponding
                to a field appearing in a TOML file.
            info (`ItemInfo`): An `ItemInfo` instance with basic info on the field.
        """
        field_desctiptor: FieldDescriptor = self._tables[info.hierarchy]._add_field(
            item=item, info=info, line_no=self._line_counter.line_no
        )
        self.field_descriptor = field_desctiptor

    def add_table(self, hierarchy: str, table: Table, info: ItemInfo) -> None:
        """
        Adds a new table to the store of tables already processed. A new key-value
        pair is added to the dictionary, where the key is the string hierarchy of the
        table, and the value is a `TableDescriptor` instance.

        Args:
            hierarchy (str): A TOML hierarchy corresponding to a table.
            table (`Table`): A Table instance, being either a tomlkit.items.Table or
                tomlkit.items.InlineTable.
            info (`ItemInfo`): An `ItemInfo` instance with basic info on the table.
        """
        table_descriptor = TableDescriptor._from_table_item(
            table=table, info=info, line_no=self._line_counter.line_no
        )
        self._tables.update({hierarchy: table_descriptor})

    def get_stylings(self, style_info: ItemInfo) -> StylingDescriptors:
        """
        Returns all stylings associated with a specific table. Stylings are separated
        by whether its whitespace or comments.

        Args:
            style_info (`ItemInfo`): An `ItemInfo` instance with basic info on the styling.

        Returns:
            `StylingDescriptors`: A `StylingDescriptors` instance containing all
                stylings associated with a table.
        """
        if item_is_table(item_type=style_info.parent_type):
            styling_positions = self._tables[style_info.hierarchy].stylings
        else:
            styling_positions = self.field_descriptor.stylings
        return styling_positions


class DescriptorStore:
    """
    A consolidated store which manages and maintains all three major stores,
    `DocumentStore`, `ArrayOfTablesStore`, and `TableStore`.

    Attributes:
        document (`DocumentStore`): A store for all fields and stylings appearing
            in the top-level space of the TOML file.
        array_of_tables (`ArrayOfTablesStore`): A store for all array of tables
            appearing within the TOML file.
        tables (`TableStore`): A store for all tables, not located in array of tables,
            appearing within the TOML file.
    """

    def __init__(self, line_counter: LineCounter) -> None:
        self._line_counter = line_counter

        # Object for storing any attributes occurring in top-level space
        self.document = DocumentStore(line_counter=self._line_counter)

        # Object for storing any array of tables objects
        self.array_of_tables = ArrayOfTablesStore(line_counter=self._line_counter)

        # Object for storing any attributes occurring within at least one table
        self.tables = TableStore(line_counter=self._line_counter)

    def _store_choice(self, info: ItemInfo) -> BaseStore:
        """
        Private method which desides which of the three stores to update.
        """
        descriptor_store: BaseStore
        item_type: Item = info.item_type
        if (
            item_type == "document"
            or not info.hierarchy
            and item_type in {"array", "field", "comment", "whitespace"}
        ):
            descriptor_store = self.document
        elif info.from_aot:
            descriptor_store = self.array_of_tables
        else:
            descriptor_store = self.tables

        return descriptor_store

    def update_styling(self, style: Stylings, style_info: ItemInfo) -> None:
        """
        Adds a new styling, corresponding to a string comment or whitespace within
        a TOML file, to a store.
        """
        descriptor_store = self._store_choice(info=style_info)
        styling_positions = descriptor_store.get_stylings(style_info=style_info)
        styling_positions._update_stylings(
            style=style, info=style_info, line_no=self._line_counter.line_no
        )

    def update_field_descriptor(self, item: items.Item, info: ItemInfo) -> None:
        """
        Adds a new field, corresponding to a non-table key-value pair within a
        TOML file, to a store.
        """
        descriptor_store = self._store_choice(info=info)
        descriptor_store.update(item=item, info=info)

    def update_array_comment(self, array: items.Array, info: ItemInfo) -> None:
        """
        Adds a comment to an array field, corresponding to a non-table key-value
        pair within a TOML file, to a store.
        """
        descriptor_store = self._store_choice(info=info)
        descriptor_store.field_descriptor._update_comment(
            item=array, line_no=self._line_counter.line_no
        )

    def update_table_descriptor(
        self, hierarchy: str, table: Table, table_info: ItemInfo
    ) -> None:
        """
        Adds a new table, corresponding to a table pair within a TOML file, to a
        store.
        """
        descriptor_store: BaseTableStore
        if table_info.from_aot:
            descriptor_store = self.array_of_tables
        else:
            descriptor_store = self.tables

        descriptor_store.add_table(hierarchy=hierarchy, table=table, info=table_info)



from tomlkit_extras._exceptions import (
    BaseTOMLError,
    HierarchyModificationError,
    InvalidArrayItemError,
    InvalidArrayOfTablesError,
    InvalidFieldError,
    InvalidHierarchyDeletionError,
    InvalidHierarchyError,
    InvalidHierarchyRetrievalError,
    InvalidHierarchyUpdateError,
    InvalidStylingError,
    InvalidTableError,
    InvalidTOMLStructureError,
    KeyNotProvidedError,
    NotContainerLikeError,
    TOMLConversionError,
    TOMLDecodingError,
    TOMLInsertionError,
    TOMLReadError,
)
from tomlkit_extras._file_validator import load_toml_file
from tomlkit_extras._hierarchy import Hierarchy
from tomlkit_extras._utils import (
    contains_out_of_order_tables,
    create_array,
    create_array_of_tables,
    create_inline_table,
    create_table,
    create_toml_document,
    safe_unwrap,
)
from tomlkit_extras.descriptor._descriptor import TOMLDocumentDescriptor
from tomlkit_extras.descriptor._descriptors import (
    AoTDescriptor,
    FieldDescriptor,
    StyleDescriptor,
    TableDescriptor,
)
from tomlkit_extras.descriptor._helpers import CommentDescriptor
from tomlkit_extras.toml._comments import get_array_field_comment, get_comments
from tomlkit_extras.toml._delete import delete_from_toml_source
from tomlkit_extras.toml._insert import (
    attribute_insert,
    container_insert,
    general_insert,
)
from tomlkit_extras.toml._out_of_order import (
    fix_out_of_order_table,
    fix_out_of_order_tables,
)
from tomlkit_extras.toml._retrieval import (
    get_attribute_from_toml_source,
    get_positions,
    is_toml_instance,
)
from tomlkit_extras.toml._update import update_toml_source

__version__ = "0.2.0"
__all__ = [
    "load_toml_file",
    "Hierarchy",
    "delete_from_toml_source",
    "TOMLDocumentDescriptor",
    "update_toml_source",
    "contains_out_of_order_tables",
    "create_array",
    "create_array_of_tables",
    "create_inline_table",
    "create_table",
    "create_toml_document",
    "safe_unwrap",
    "get_array_field_comment",
    "get_comments",
    "StructureComment",
    "CommentDescriptor",
    "AoTDescriptor",
    "FieldDescriptor",
    "StyleDescriptor",
    "TableDescriptor",
    "attribute_insert",
    "container_insert",
    "general_insert",
    "fix_out_of_order_table",
    "fix_out_of_order_tables",
    "get_attribute_from_toml_source",
    "get_positions",
    "is_toml_instance",
    "BaseTOMLError",
    "HierarchyModificationError",
    "InvalidArrayItemError",
    "InvalidArrayOfTablesError",
    "InvalidHierarchyError",
    "InvalidFieldError",
    "InvalidHierarchyDeletionError",
    "InvalidHierarchyRetrievalError",
    "InvalidHierarchyUpdateError",
    "InvalidStylingError",
    "InvalidTableError",
    "InvalidTOMLStructureError",
    "KeyNotProvidedError",
    "NotContainerLikeError",
    "TOMLConversionError",
    "TOMLDecodingError",
    "TOMLInsertionError",
    "TOMLReadError",
]



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional, Set, Union

from tomlkit import TOMLDocument, items
from tomlkit.container import OutOfOrderTableProxy

from tomlkit_extras._hierarchy import Hierarchy
from tomlkit_extras._typing import Item, TOMLValidReturn


@dataclass(frozen=True)
class CommentDescriptor:
    """
    A dataclass which provides detail for a comment that is directly
    associated with a particular field or table.

    Attributes:
        comment (str): A string representing the comment.
        line_no (int): An integer line number where the comment is located.
    """

    comment: str
    line_no: int


@dataclass
class LineCounter:
    """
    Line counter to keep track of the number of lines seen while
    traversing through and mapping a TOML file within the
    `TOMLDocumentDescriptor` class.
    """

    line_no: int = 0

    def add_lines(self, lines: int) -> None:
        """Add a custom number of lines."""
        self.line_no += lines

    def add_line(self) -> None:
        """Add one line."""
        self.line_no += 1

    def reset_line_no(self) -> None:
        """Reset the line number to 0."""
        self.line_no = 0


def create_comment_descriptor(
    item: items.Item, line_no: Optional[int]
) -> Optional[CommentDescriptor]:
    """
    A private function that creates a `CommentDescriptor` instance which
    provides detail for a comment that is directly associated with a
    particular field or table.

    Can return None if there is no line number corresponding to the item,
    indicating that there is no comment.
    """
    return (
        CommentDescriptor(comment=item.trivia.comment, line_no=line_no)
        if line_no is not None
        else None
    )


def item_is_table(item_type: Optional[Item]) -> bool:
    """
    A private function that determines if an `Item`, which is a literal that
    identifies the type of `tomlkit` object, is a table.
    """
    return item_type in {"table", "inline-table"}


def find_child_tables(root_hierarchy: str, hierarchies: Set[str]) -> Set[str]:
    """
    A private function that outputs all child hierarchies of a specific root
    hierarchy, based on a set of string hierarchies.
    """
    children_hierarchies: Set[str] = set()

    root_hierarchy_obj = Hierarchy.from_str_hierarchy(hierarchy=root_hierarchy)

    for hierarchy in hierarchies:
        if root_hierarchy_obj.is_child_hierarchy(hierarchy=hierarchy):
            children_hierarchies.add(hierarchy)

    return children_hierarchies


def get_item_type(toml_item: Union[TOMLDocument, TOMLValidReturn]) -> Item:
    """
    A private function that will return an `Item`, corresponding to a string
    literal representing the type of the TOML structure, given a
    `tomlkit.TOMLDocument` or `TOMLValidReturn` instance

    So, for example, if a tomlkit.TOMLDocument is passed in, then 'document'
    would be returned. If a `tomlkit.items.AoT` is passed, then 'array-of-tables'
    would output.
    """
    toml_item_type: Item

    if isinstance(toml_item, TOMLDocument):
        toml_item_type = "document"
    elif isinstance(toml_item, items.Table):
        toml_item_type = "super-table" if toml_item.is_super_table() else "table"
    elif isinstance(toml_item, OutOfOrderTableProxy):
        toml_item_type = "table"
    elif isinstance(toml_item, items.InlineTable):
        toml_item_type = "inline-table"
    elif isinstance(toml_item, items.Comment):
        toml_item_type = "comment"
    elif isinstance(toml_item, items.Whitespace):
        toml_item_type = "whitespace"
    elif isinstance(toml_item, items.AoT):
        toml_item_type = "array-of-tables"
    elif isinstance(toml_item, items.Array):
        toml_item_type = "array"
    else:
        toml_item_type = "field"

    return toml_item_type



from __future__ import annotations

from typing import TYPE_CHECKING, Any, List, Optional, Set, Tuple

if TYPE_CHECKING:
    from tomlkit_extras._typing import TOMLHierarchy


def standardize_hierarchy(hierarchy: TOMLHierarchy) -> Hierarchy:
    """
    Accepts a `TOMLHierarchy` instance, being an instance of string or `Hierarchy`,
    and returns a `Hierarchy` instance.

    Args:
        hierarchy (`TOMLHierarchy`) A `TOMLHierarchy` instance.

    Returns:
        `Hierarchy`: A `Hierarchy` instance.
    """
    if isinstance(hierarchy, str):
        hierarchy_final = Hierarchy.from_str_hierarchy(hierarchy=hierarchy)
    else:
        hierarchy_final = hierarchy
    return hierarchy_final


class Hierarchy:
    """
    A class representing a hierarchy, commonly used for managing and manipulating
    string-based hierarchical structures, such as those found in TOML files.

    This class provides methods to construct, decompose, compare, and navigate
    hierarchical structures. It also supports operations to retrieve ancestors,
    append levels, and determine relationships between hierarchies.

    Attributes:
        hierarchy (Tuple[str, ...]): A tuple representing the base levels of the hierarchy.
        attribute (str): The final level or attribute of the hierarchy.
    """

    def __init__(self, hierarchy: Tuple[str, ...], attribute: str) -> None:
        self.hierarchy = hierarchy
        self.attribute = attribute

    def __eq__(self, hierarchy: Any) -> bool:
        if not isinstance(hierarchy, (str, Hierarchy)):
            raise TypeError(
                "Expected an instance of string or Hierarchy,"
                f" but got {type(hierarchy).__name__}"
            )

        hierarchy_arg = standardize_hierarchy(hierarchy=hierarchy)
        return (
            self.hierarchy == hierarchy_arg.hierarchy
            and self.attribute == hierarchy_arg.attribute
        )

    def __str__(self) -> str:
        return self.full_hierarchy_str

    def __repr__(self) -> str:
        return f"<Hierarchy {self.full_hierarchy_str}>"

    def __len__(self) -> int:
        return self.depth

    @staticmethod
    def parent_hierarchy(hierarchy: str) -> str:
        """
        A static method which returns the parent hierarchy of a string
        representation of a TOML hierarchy.

        Args:
            hierarchy (str): A string representation of a TOML hierarchy.

        Returns:
            str: A string instance representing the parent TOML hierarchy of the
                hierarchy passed in.
        """
        return ".".join(hierarchy.split(".")[:-1])

    @staticmethod
    def create_hierarchy(hierarchy: str, attribute: str) -> str:
        """
        A static method which creates a string representation of a TOML hierarchy
        from a string TOML hierarchy and an update string instance.

        Args:
            hierarchy (str) A string instance representing a TOML hierarchy.
            attribute (str): An attribute to be added to the existing hierarchy.

        Returns:
            str: A string instance representing an updated TOML hierarchy.
        """
        full_hierarchy = str()

        if hierarchy:
            full_hierarchy += hierarchy

        if attribute and not full_hierarchy:
            full_hierarchy += attribute
        elif attribute:
            full_hierarchy += "." + attribute

        return full_hierarchy

    @classmethod
    def from_str_hierarchy(cls, hierarchy: str) -> Hierarchy:
        """
        A class method which returns a `Hierarchy` instance from a string instance of
        a TOML hierarchy.

        Args:
            hierarchy (str) A string instance representing a TOML hierarchy.

        Returns:
            `Hierarchy`: A `Hierarchy` instance.
        """
        hierarchy_decomposed = hierarchy.split(".")
        assert len(hierarchy_decomposed) > 0

        return cls.from_list_hierarchy(hierarchy=hierarchy_decomposed)

    @classmethod
    def from_list_hierarchy(cls, hierarchy: List[str]) -> Hierarchy:
        """
        A class method which returns a `Hierarchy` instance from a list instance of
        strings representing each individual level in a TOML hierarchy.

        Args:
            hierarchy (List[str]) A list instance of strings representing each
                individual level in a TOML hierarchy

        Returns:
            `Hierarchy`: A `Hierarchy` instance.
        """
        if not hierarchy:
            raise ValueError("There must be an existing hierarchy")

        attribute_hierarchy: Tuple[str, ...]

        # Compile attribute and rest of hierarchy
        if len(hierarchy) == 1:
            attribute = hierarchy[0]
            attribute_hierarchy = tuple()
        else:
            attribute = hierarchy[-1]
            attribute_hierarchy = tuple(hierarchy[:-1])

        return cls(attribute=attribute, hierarchy=attribute_hierarchy)

    @property
    def depth(self) -> int:
        """
        Returns the depth of the hierarchy, also known as the number of levels
        in the hierarchy.
        """
        return len(self.full_hierarchy)

    @property
    def root_attribute(self) -> str:
        """Returns the root/first level of the hierarchy."""
        if not self.hierarchy:
            return self.attribute
        else:
            return self.hierarchy[0]

    @property
    def full_hierarchy(self) -> Tuple[str, ...]:
        """Returns a tuple instance of the entire hierarchy."""
        return tuple(list(self.hierarchy) + [self.attribute])

    @property
    def base_hierarchy_str(self) -> str:
        """
        Returns a string instance of the entire hierarchy minus the attribute/last
        level.
        """
        return ".".join(self.hierarchy)

    @property
    def full_hierarchy_str(self) -> str:
        """Returns a string instance of the entire hierarchy."""
        if not self.base_hierarchy_str:
            return self.attribute
        else:
            return ".".join(self.full_hierarchy)

    @property
    def ancestor_hierarchies(self) -> List[str]:
        """
        Returns a list of strings representing all ancestor hierarchies of the current
        hierarchy.
        """
        sub_hierarchies: List[str] = []

        start_hierarchy = str()
        for hierarchy in self.full_hierarchy:
            start_hierarchy = Hierarchy.create_hierarchy(
                hierarchy=start_hierarchy, attribute=hierarchy
            )
            sub_hierarchies.append(start_hierarchy)

        return sub_hierarchies

    def shortest_ancestor_hierarchy(self, hierarchies: Set[str]) -> Optional[str]:
        """
        Returns the shortest hierarchy appearing in a set of string instances being
        an ancestor of the current hierarchy.

        Will return None if no hierarchy found in the set is an ancestor of the
        current hierarchy.

        Args:
            hierarchies (Set[str]): A set of strings representing TOML hierarchies.

        Returns:
            str | None: A string instance or None.
        """
        ancestor_hierarchies = sorted(self.ancestor_hierarchies, key=lambda x: len(x))
        return self._ancestor_hierarchy_match(
            ancestor_hierarchies=ancestor_hierarchies, hierarchies=hierarchies
        )

    def longest_ancestor_hierarchy(self, hierarchies: Set[str]) -> Optional[str]:
        """
        Returns the longest hierarchy appearing in a set of string instances being
        an ancestor of the current hierarchy.

        Will return None if no hierarchy found in the set is an ancestor of the
        current hierarchy.

        Args:
            hierarchies (Set[str]): A set of strings representing TOML hierarchies.

        Returns:
            str | None: A string instance or None.
        """
        ancestor_hierarchies = sorted(self.ancestor_hierarchies, key=lambda x: -len(x))
        return self._ancestor_hierarchy_match(
            ancestor_hierarchies=ancestor_hierarchies, hierarchies=hierarchies
        )

    def _ancestor_hierarchy_match(
        self, ancestor_hierarchies: List[str], hierarchies: Set[str]
    ) -> Optional[str]:
        """
        A private method that returns the first ancestor hierarchy appearing in a
        set of other string hierarchies.

        Will return None if no ancestor hierarchy was found.
        """
        for hierarchy in ancestor_hierarchies:
            if hierarchy in hierarchies:
                return hierarchy

        return None

    def add_to_hierarchy(self, update: str) -> None:
        """
        Appends at least one more level to the existing hierarchy, which updates both
        the hierarchy and attribute data fields.

        Args:
            update (str): A string instance to append to the existing hierarchy.
        """
        if update:
            update_decomposed: List[str] = update.split(".")

            attribute = update_decomposed[-1]
            if self.full_hierarchy_str:
                hierarchy_new = list(self.full_hierarchy)
            else:
                hierarchy_new = list()

            if len(update_decomposed) > 1:
                hierarchy_new += update_decomposed[:-1]

            self.hierarchy = tuple(hierarchy_new)
            self.attribute = attribute

    def is_child_hierarchy(self, hierarchy: str) -> bool:
        """
        Returns a boolean indicating whether the hierarchy passed in as an argument
        is a child of the current hierarchy.

        Args:
            hierarchy (str): A string representation of a TOML hierarchy.

        Returns:
            bool: A boolean indicating whether it is a child hierarchy of the current
                hierarchy.
        """
        parent_hierarchy = Hierarchy.parent_hierarchy(hierarchy=hierarchy)
        return self == parent_hierarchy
